from typing import TYPE_CHECKING, List, Optional, Union

from dis_snek.models.discord_objects.message import process_message_payload
from dis_snek.models.enums import MessageFlags

if TYPE_CHECKING:
    from io import IOBase
    from pathlib import Path

    from aiohttp.formdata import FormData

    from dis_snek.client import Snake
    from dis_snek.models.discord_objects.components import BaseComponent
    from dis_snek.models.discord_objects.embed import Embed
    from dis_snek.models.discord_objects.message import AllowedMentions, Message, MessageReference
    from dis_snek.models.discord_objects.sticker import Sticker
    from dis_snek.models.snowflake import Snowflake_Type


class SendMixin:
    _client: "Snake"

    async def _send_http_request(self, message_payload: Union[dict, "FormData"]) -> dict:
        raise NotImplementedError

    async def send(
        self,
        content: Optional[str] = None,
        embeds: Optional[Union[List[Union["Embed", dict]], Union["Embed", dict]]] = None,
        components: Optional[
            Union[List[List[Union["BaseComponent", dict]]], List[Union["BaseComponent", dict]], "BaseComponent", dict]
        ] = None,
        stickers: Optional[Union[List[Union["Sticker", "Snowflake_Type"]], "Sticker", "Snowflake_Type"]] = None,
        allowed_mentions: Optional[Union["AllowedMentions", dict]] = None,
        reply_to: Optional[Union["MessageReference", "Message", dict, "Snowflake_Type"]] = None,
        file: Optional[Union["IOBase", "Path", str]] = None,
        tts: bool = False,
        flags: Optional[Union[int, "MessageFlags"]] = None,
        **kwargs
    ) -> "Message":
        """
        Send a message.

        parameters:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            components: The components to include with the message.
            stickers: IDs of up to 3 stickers in the server to send in the message.
            allowed_mentions: Allowed mentions for the message.
            reply_to: Message to reference, must be from the same channel.
            filepath: Location of file to send, defaults to None.
            tts: Should this message use Text To Speech.
            flags: Message flags to apply.

        returns:
            New message object that was sent.
        """
        message_payload = process_message_payload(
            content=content,
            embeds=embeds,
            components=components,
            stickers=stickers,
            allowed_mentions=allowed_mentions,
            reply_to=reply_to,
            file=file,
            tts=tts,
            flags=flags,
        )
        if kwargs:
            for key, value in kwargs.items():
                message_payload[key] = value

        message_data = await self._send_http_request(message_payload)
        if message_data:
            return self._client.cache.place_message_data(message_data)



from enum import IntEnum
from typing import Optional, TYPE_CHECKING, Union, Dict, Any, List

import attr
from aiohttp import FormData

from dis_snek.mixins.send import SendMixin
from dis_snek.models.discord import DiscordObject
from dis_snek.models.snowflake import to_snowflake
from dis_snek.utils.attr_utils import define
from dis_snek.utils.input_utils import _bytes_to_base64_data

if TYPE_CHECKING:
    from io import IOBase
    from pathlib import Path

    from dis_snek.client import Snake
    from dis_snek.models.snowflake import Snowflake_Type
    from dis_snek.models.discord_objects.channel import TYPE_MESSAGEABLE_CHANNEL
    from dis_snek.models.discord_objects.components import BaseComponent
    from dis_snek.models.discord_objects.embed import Embed
    from dis_snek.models.discord_objects.message import AllowedMentions, Message, MessageReference
    from dis_snek.models.discord_objects.sticker import Sticker


class WebhookTypes(IntEnum):
    INCOMING = 1
    """Incoming Webhooks can post messages to channels with a generated token"""
    CHANNEL_FOLLOWER = 2
    """Channel Follower Webhooks are internal webhooks used with Channel Following to post new messages into channels"""
    APPLICATION = 3
    """Application webhooks are webhooks used with Interactions"""


@define
class Webhook(DiscordObject, SendMixin):
    type: WebhookTypes = attr.ib()
    """The type of webhook"""

    application_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """the bot/OAuth2 application that created this webhook"""

    guild_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """the guild id this webhook is for, if any"""
    channel_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """the channel id this webhook is for, if any"""
    user_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """the user this webhook was created by"""

    name: Optional[str] = attr.ib(default=None)
    """the default name of the webhook"""
    avatar: Optional[str] = attr.ib(default=None)
    """the default user avatar hash of the webhook"""
    token: str = attr.ib()
    """the secure token of the webhook (returned for Incoming Webhooks)"""
    url: Optional[str] = attr.ib(default=None)
    """the url used for executing the webhook (returned by the webhooks OAuth2 flow)"""

    source_guild_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """the guild of the channel that this webhook is following (returned for Channel Follower Webhooks)"""
    source_channel_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """the channel that this webhook is following (returned for Channel Follower Webhooks)"""

    @classmethod
    async def create(
        cls,
        client: "Snake",
        channel: Union["Snowflake_Type", "TYPE_MESSAGEABLE_CHANNEL"],
        name: str,
        avatar: Optional[bytes],
    ):
        if name.lower() == "clyde":
            raise ValueError('Webhook names cannot be "Clyde"')

        if not isinstance(channel, (str, int)):
            channel = to_snowflake(channel)

        if avatar:
            avatar = _bytes_to_base64_data(avatar)

        data = await client.http.create_webhook(channel, name, avatar)

        new_cls = cls.from_dict(data, client)

        return new_cls

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        if data.get("user"):
            user = client.cache.place_user_data(data.pop("user"))
            data["user_id"] = user.id
        return data

    async def delete(self) -> None:
        """Delete this webhook"""
        await self._client.http.delete_webhook(self.id, self.token)

    async def _send_http_request(self, message_payload: Union[dict, "FormData"]) -> dict:
        return await self._client.http.execute_webhook(self.id, self.token, message_payload)

    async def send(
        self,
        content: Optional[str] = None,
        embeds: Optional[Union[List[Union["Embed", dict]], Union["Embed", dict]]] = None,
        components: Optional[
            Union[List[List[Union["BaseComponent", dict]]], List[Union["BaseComponent", dict]], "BaseComponent", dict]
        ] = None,
        stickers: Optional[Union[List[Union["Sticker", "Snowflake_Type"]], "Sticker", "Snowflake_Type"]] = None,
        allowed_mentions: Optional[Union["AllowedMentions", dict]] = None,
        reply_to: Optional[Union["MessageReference", "Message", dict, "Snowflake_Type"]] = None,
        file: Optional[Union["IOBase", "Path", str]] = None,
        tts: bool = False,
        flags: Optional[Union[int, "MessageFlags"]] = None,
        username: str = None,
        avatar_url: str = None,
    ):
        """
        Send a message as this webhook

        parameters:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            components: The components to include with the message.
            stickers: IDs of up to 3 stickers in the server to send in the message.
            allowed_mentions: Allowed mentions for the message.
            reply_to: Message to reference, must be from the same channel.
            filepath: Location of file to send, defaults to None.
            tts: Should this message use Text To Speech.
            flags: Message flags to apply.
            username: The username to use
            avatar_url: The url of an image to use as the avatar

        returns:
            New message object that was sent.
        """
        return await super().send(
            content,
            embeds,
            components,
            stickers,
            allowed_mentions,
            reply_to,
            file,
            tts,
            flags,
            username=username,
            avatar_url=avatar_url,
        )

    # todo: edit webhook messaage



from typing import TYPE_CHECKING, Any, List

from dis_snek.const import MISSING
from dis_snek.models.route import Route

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type


class MessageRequests:
    request: Any

    async def create_message(self, payload: dict, channel_id: "Snowflake_Type") -> dict:
        """Send a message to the specified channel."""
        return await self.request(Route("POST", f"/channels/{channel_id}/messages"), data=payload)

    async def delete_message(
        self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type", reason: str = MISSING
    ) -> Any:
        """Deletes a message from the specified channel. Incomplete."""
        await self.request(Route("DELETE", f"/channels/{channel_id}/messages/{message_id}"), reason=reason)

    async def bulk_delete_messages(
        self, channel_id: "Snowflake_Type", message_ids: List["Snowflake_Type"], reason: str = MISSING
    ) -> None:
        """
        Delete multiple messages in a single request.

        parameters:
            channel_id: The id of the channel these messages are in
            message_ids: A list of message ids to delete
            reason: The reason for this action
        """
        return await self.request(
            Route("POST", f"/channels/{channel_id}/messages/bulk-delete"), data={"messages": message_ids}, reason=reason
        )

    async def get_message(self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type") -> dict:
        """
        Get a specific message in the channel. Returns a message object on success.


        parameters:
            channel_id: the channel this message belongs to
            message_id: the id of the message
        returns:
            message or None
        """
        return await self.request(Route("GET", f"/channels/{channel_id}/messages/{message_id}"))

    async def pin_message(self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type") -> None:
        """Pin a message to a channel

        parameters:
            channel_id: Channel to pin message to
            message_id: Message to pin
        """
        return await self.request(Route("PUT", f"/channels/{channel_id}/pins/{message_id}"))

    async def unpin_message(self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type") -> None:
        """Unpin a message to a channel

        parameters:
            channel_id: Channel to unpin message to
            message_id: Message to unpin
        """
        return await self.request(Route("DELETE", f"/channels/{channel_id}/pins/{message_id}"))

    async def edit_message(
        self,
        payload: dict,
        channel_id: "Snowflake_Type",
        message_id: "Snowflake_Type",
    ) -> dict:
        """Edit an existing message

        parameters:
            payload:
            channel_id: Channel of message to edit.
            message_id: Message to edit.

        returns:
            Message object of edited message
        """
        return await self.request(Route("PATCH", f"/channels/{channel_id}/messages/{message_id}"), data=payload)

    async def crosspost_message(self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type") -> dict:
        """
        Crosspost a message in a News Channel to following channels.

        parameters:
            channel_id: Channel the message is in
            message_id: The id of the message to crosspost
        returns:
            message object
        """
        return await self.request(Route("POST", f"/channels/{channel_id}/messages/{message_id}/crosspost"))



from typing import TYPE_CHECKING, List, Dict, Any, Optional

import attr

from dis_snek.errors import NotFound, Forbidden
from dis_snek.models.discord_objects.channel import BaseChannel
from dis_snek.models.discord_objects.guild import Guild
from dis_snek.models.discord_objects.message import Message
from dis_snek.models.discord_objects.role import Role
from dis_snek.models.discord_objects.user import Member, User
from dis_snek.models.snowflake import to_snowflake
from dis_snek.utils.attr_utils import field
from dis_snek.utils.cache import TTLCache

if TYPE_CHECKING:
    from dis_snek.client import Snake
    from dis_snek.models.discord_objects.channel import DM, TYPE_ALL_CHANNEL
    from dis_snek.models.snowflake import Snowflake_Type


@attr.define()
class GlobalCache:
    _client: "Snake" = field()

    # Non expiring discord objects cache
    user_cache: dict = field(factory=dict)  # key: user_id
    member_cache: dict = field(factory=dict)  # key: (guild_id, user_id)
    channel_cache: dict = field(factory=dict)  # key: channel_id
    guild_cache: dict = field(factory=dict)  # key: guild_id

    # Expiring discord objects cache
    message_cache: TTLCache = field(factory=TTLCache)  # key: (channel_id, message_id)
    role_cache: TTLCache = field(factory=TTLCache)  # key: role_id

    # Expiring id reference cache
    dm_channels: TTLCache = field(factory=TTLCache)  # key: user_id
    user_guilds: TTLCache = field(factory=TTLCache)  # key: user_id; value: set[guild_id]

    # User cache methods

    async def get_user(self, user_id: "Snowflake_Type", request_fallback=True) -> User:
        user_id = to_snowflake(user_id)

        user = self.user_cache.get(user_id)
        if request_fallback and user is None:
            data = await self._client.http.get_user(user_id)
            user = self.place_user_data(data)
        return user

    def place_user_data(self, data) -> User:
        user_id = to_snowflake(data["id"])

        user = self.user_cache.get(user_id)

        if user is None:
            user = User.from_dict(data, self._client)
            self.user_cache[user_id] = user
        else:
            user.update_from_dict(data)
        return user

    # Member cache methods

    async def get_member(self, guild_id: "Snowflake_Type", user_id: "Snowflake_Type", request_fallback=True) -> Member:
        guild_id = to_snowflake(guild_id)
        user_id = to_snowflake(user_id)
        member = self.member_cache.get((guild_id, user_id))
        if request_fallback and member is None:
            data = await self._client.http.get_member(guild_id, user_id)
            member = self.place_member_data(guild_id, data)
        return member

    def place_member_data(self, guild_id, data) -> Member:
        guild_id = to_snowflake(guild_id)
        is_user = "member" in data
        user_id = to_snowflake(data["id"] if is_user else data["user"]["id"])

        member = self.member_cache.get((guild_id, user_id))
        if member is None:
            member_extra = {"guild_id": guild_id}
            member = data["member"] if is_user else data
            member.update(member_extra)

            member = Member.from_dict(data, self._client)
            self.member_cache[(guild_id, user_id)] = member
        else:
            member.update_from_dict(data)

        self.place_user_guild(user_id, guild_id)
        return member

    def place_user_guild(self, user_id: "Snowflake_Type", guild_id: "Snowflake_Type"):
        user_id = to_snowflake(user_id)
        guild_id = to_snowflake(guild_id)
        if user_id == self._client.user.id:
            # noinspection PyProtectedMember
            self._client.user._add_guilds({guild_id})
        else:
            guilds = self.user_guilds.get(user_id)
            if guilds:
                guilds.add(guild_id)

    async def is_user_in_guild(self, user_id: "Snowflake_Type", guild_id: "Snowflake_Type", request_fallback=True):
        user_id = to_snowflake(user_id)
        guild_id = to_snowflake(guild_id)

        # Try to get guild members list from the cache, without sending requests
        guild = await self.get_guild(guild_id, request_fallback=False)
        if guild and (user_id in guild.members.ids):
            return True
        # If no such guild in cache or member not in guild cache, try to get member directly. May send requests
        try:
            member = await self.get_member(guild_id, user_id, request_fallback)
        except (NotFound, Forbidden):  # there is no such member in the guild (as per request)
            pass
        else:
            if member:
                return True

        return False

    async def get_user_guild_ids(self, user_id: "Snowflake_Type", calculation_fallback=True, request_fallback=True):
        user_id = to_snowflake(user_id)
        guild_ids = self.user_guilds.get(user_id)
        if not guild_ids and calculation_fallback:
            guild_ids = [
                guild_id
                for guild_id in self._client.user.guilds.ids
                if await self.is_user_in_guild(user_id, guild_id, request_fallback)
            ]
        return guild_ids

    # Message cache methods

    async def get_message(
        self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type", request_fallback=True
    ) -> Message:
        channel_id = to_snowflake(channel_id)
        message_id = to_snowflake(message_id)
        message = self.message_cache.get((channel_id, message_id))
        if request_fallback and message is None:
            data = await self._client.http.get_message(channel_id, message_id)
            message = self.place_message_data(data)
        return message

    def place_message_data(self, data) -> Message:
        channel_id = to_snowflake(data["channel_id"])
        message_id = to_snowflake(data["id"])
        message = self.message_cache.get((channel_id, message_id))
        if message is None:
            message = Message.from_dict(data, self._client)
            self.message_cache[(channel_id, message_id)] = message
        else:
            message.update_from_dict(data)
        return message

    # Channel cache methods

    async def get_channel(self, channel_id: "Snowflake_Type", request_fallback=True) -> "TYPE_ALL_CHANNEL":
        channel_id = to_snowflake(channel_id)
        channel = self.channel_cache.get(channel_id)
        if request_fallback and channel is None:
            data = await self._client.http.get_channel(channel_id)
            channel = self.place_channel_data(data)
        return channel

    def place_channel_data(self, data) -> "TYPE_ALL_CHANNEL":
        channel_id = to_snowflake(data["id"])
        channel = self.channel_cache.get(channel_id)
        if channel is None:
            channel = BaseChannel.from_dict_factory(data, self._client)
            self.channel_cache[channel_id] = channel
        else:
            channel.update_from_dict(data)
        return channel

    def place_dm_channel_id(self, user_id, channel_id):
        self.dm_channels[to_snowflake(user_id)] = to_snowflake(channel_id)

    async def get_dm_channel_id(self, user_id):
        user_id = to_snowflake(user_id)
        channel_id = self.dm_channels.get(user_id)
        if channel_id is None:
            data = await self._client.http.create_dm(user_id)
            channel = self.place_channel_data(data)
            channel_id = channel.id
        return channel_id

    async def get_dm_channel(self, user_id) -> "DM":
        user_id = to_snowflake(user_id)
        channel_id = await self.get_dm_channel_id(user_id)
        channel = await self.get_channel(channel_id)
        return channel

    # Guild cache methods

    async def get_guild(self, guild_id: "Snowflake_Type", request_fallback=True) -> Guild:
        guild_id = to_snowflake(guild_id)
        guild = self.guild_cache.get(guild_id)
        if request_fallback and guild is None:
            data = await self._client.http.get_guild(guild_id)
            guild = self.place_guild_data(data)
        return guild

    def place_guild_data(self, data) -> Guild:
        guild_id = to_snowflake(data["id"])
        guild = self.guild_cache.get(guild_id)
        if guild is None:
            guild = Guild.from_dict(data, self._client)
            self.guild_cache[guild_id] = guild
        else:
            guild.update_from_dict(data)
        return guild

    # Roles cache methods

    async def get_role(
        self, guild_id: "Snowflake_Type", role_id: "Snowflake_Type", request_fallback=True
    ) -> Optional[Role]:
        guild_id = to_snowflake(guild_id)
        role_id = to_snowflake(role_id)
        role = self.role_cache.get(role_id)
        if request_fallback and role is None:
            data = await self._client.http.get_roles(guild_id)
            try:
                role = self.place_role_data(guild_id, data)[role_id]
            except KeyError:
                return None
        return role

    def place_role_data(
        self, guild_id: "Snowflake_Type", data: List[Dict["Snowflake_Type", Any]]
    ) -> Dict["Snowflake_Type", Role]:
        guild_id = to_snowflake(guild_id)

        roles = {}
        for role_data in data:  # todo not update cache expiration order for roles
            role_data.update({"guild_id": guild_id})
            role_id = to_snowflake(role_data["id"])

            role = self.role_cache.get(role_id)
            if role is None:
                role = Role.from_dict(role_data, self._client)
                self.role_cache[role_id] = role
            else:
                role.update_from_dict(role_data)

            roles[role_id] = role

        return roles



from typing import TYPE_CHECKING, Optional, Union

import attr

if TYPE_CHECKING:
    from os import PathLike

    from dis_snek.client import Snake


@attr.s(slots=True)
class Asset:
    """
    Represents a discord asset.

    Attributes:
        BASE str: The `cdn` address for assets
        url str: The URL of this asset
        hash Optional[str]: The hash of this asset
    """

    BASE = "https://cdn.discordapp.com"

    _client: "Snake" = attr.field()
    url: str = attr.field()
    hash: Optional[str] = attr.field(default=None)

    @classmethod
    def from_path_hash(cls, client: "Snake", path: str, asset_hash: str) -> "Asset":
        url = f"{cls.BASE}/{path.format(asset_hash)}"
        return cls(client=client, url=url, hash=asset_hash)

    @property
    def animated(self) -> bool:
        """True if this asset is animated"""
        if not self.hash:
            return None
        return self.hash.startswith("a_")

    async def get(self, extension: Optional[str] = None, size: Optional[int] = None) -> bytes:
        if not extension:
            extension = ".gif" if self.animated else ".png"

        url = self.url

        if size:
            if not ((size != 0) and (size & (size - 1) == 0)):  # if not power of 2
                raise ValueError("Size should be a power of 2")
            if not 16 <= size <= 4096:
                raise ValueError("Size should be between 16 and 4096")

            url = f"{url}?size={size}"

        url = url + extension
        return await self._client.http.request_cdn(url, self)

    async def save(
        self, fd: Union[str, bytes, "PathLike", int], extension: Optional[str] = None, size: Optional[int] = None
    ) -> int:
        content = await self.get(extension=extension, size=size)
        with open(fd, "wb") as f:
            return f.write(content)



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import attr
from attr.converters import optional

from dis_snek.mixins.serialization import DictSerializationMixin
from dis_snek.models.snowflake import SnowflakeObject, to_snowflake
from dis_snek.utils.attr_utils import define, field
from dis_snek.utils.converters import list_converter
from dis_snek.utils.serializer import dict_filter_none, no_export_meta

if TYPE_CHECKING:
    from dis_snek.client import Snake
    from dis_snek.models.discord_objects.guild import Guild
    from dis_snek.models.discord_objects.user import User
    from dis_snek.models.discord_objects.role import Role
    from dis_snek.models.snowflake import Snowflake_Type


@define()
class Emoji(SnowflakeObject, DictSerializationMixin):
    """Represent a basic emoji used in discord."""

    id: Optional["Snowflake_Type"] = attr.ib(
        default=None, converter=optional(to_snowflake)
    )  # can be None for Standard Emoji
    """The custom emoji id. Leave empty if you are using standard unicode emoji."""
    name: Optional[str] = attr.ib(default=None)
    """The custom emoji name, or standard unicode emoji in string"""
    animated: bool = attr.ib(default=False)
    """Whether this emoji is animated"""

    @classmethod
    def unicode(cls, emoji: str):
        return cls(name=emoji)

    def __str__(self) -> str:
        return f"<{'a:' if self.animated else ''}{self.name}:{self.id}>"  # <:thinksmart:623335224318754826>

    @property
    def req_format(self) -> str:
        """
        Format used for web request.
        """
        if self.id:
            return f"{self.name}:{self.id}"
        else:
            return self.name


@define()
class CustomEmoji(Emoji):
    """Represent a custom emoji in a guild with all its properties."""

    _client: "Snake" = field(metadata=no_export_meta)

    require_colons: bool = attr.ib(default=False)
    """Whether this emoji must be wrapped in colons"""
    managed: bool = attr.ib(default=False)
    """Whether this emoji is managed"""
    available: bool = attr.ib(default=False)
    """Whether this emoji can be used, may be false due to loss of Server Boosts."""

    _creator_id: Optional["Snowflake_Type"] = attr.ib(default=None, converter=optional(to_snowflake))
    _role_ids: List["Snowflake_Type"] = attr.ib(factory=list, converter=optional(list_converter(to_snowflake)))
    _guild_id: Optional["Snowflake_Type"] = attr.ib(default=None, converter=optional(to_snowflake))

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        creator_dict = data.pop("user", None)
        data["creator_id"] = client.cache.place_user_data(creator_dict).id if creator_dict else None

        if "roles" in data:
            data["role_ids"] = data.pop("roles")

        return data

    @classmethod
    def from_dict(cls, data: Dict[str, Any], client: "Snake") -> "CustomEmoji":
        data = cls._process_dict(data, client)
        return cls(client=client, **cls._filter_kwargs(data, cls._get_init_keys()))

    @property
    def is_usable(self) -> bool:
        """
        Determines if this emoji is usable by the current user.
        """
        if not self.available:
            return False
        # todo: check roles
        return True

    async def get_creator(self) -> "User":
        """
        Get the user who created this emoji

        Returns:
            User object
        """
        return await self._client.cache.get_user(self._creator_id)

    async def get_roles(self) -> List["Role"]:
        """
        Gets the roles allowed to use this emoji

        Returns:
            List of roles
        """
        return [await self._client.cache.get_role(self._guild_id, r_id) for r_id in self._role_ids]

    async def get_guild(self) -> "Guild":
        """
        Get the guild associated with this emoji

        Returns:
            Guild object
        """
        return await self._client.cache.get_guild(self._guild_id)

    async def edit(
        self,
        name: Optional[str] = None,
        roles: Optional[List[Union["Snowflake_Type", "Role"]]] = None,
        reason: Optional[str] = None,
    ) -> "CustomEmoji":
        """
        Modify the custom emoji information.

        parameters:
            name: The name of the emoji.
            roles: The roles allowed to use this emoji.
            reason: Attach a reason to this action, used for audit logs.

        returns:
            The newly modified custom emoji.
        """
        data_payload = dict_filter_none(
            dict(
                name=name,
                roles=roles,
            )
        )

        updated_data = await self._client.http.modify_guild_emoji(data_payload, self._guild_id, self.id, reason=reason)
        self.update_from_dict(updated_data)
        return self

    async def delete(self, reason: Optional[str] = None) -> None:
        """
        Deletes the custom emoji from the guild.

        parameters:
            reason: Attach a reason to this action, used for audit logs.
        """
        if not self._guild_id:
            raise ValueError("Cannot delete emoji, no guild id set.")

        await self._client.http.delete_guild_emoji(self._guild_id, self.id, reason=reason)


def process_emoji_req_format(emoji: Optional[Union[Emoji, dict, str]]) -> Optional[str]:
    if not emoji:
        return emoji

    if isinstance(emoji, str):
        return emoji

    if isinstance(emoji, dict):
        emoji = Emoji.from_dict(emoji)

    if isinstance(emoji, Emoji):
        return emoji.req_format

    raise ValueError(f"Invalid emoji: {emoji}")


def process_emoji(emoji: Optional[Union[Emoji, dict, str]]) -> Optional[dict]:
    if not emoji:
        return emoji

    if isinstance(emoji, dict):
        return emoji

    if isinstance(emoji, str):
        emoji = Emoji.unicode(emoji)

    if isinstance(emoji, Emoji):
        return emoji.to_dict()

    raise ValueError(f"Invalid emoji: {emoji}")



from typing import TYPE_CHECKING, List

from dis_snek.const import MISSING
from dis_snek.models.route import Route

if TYPE_CHECKING:
    from dis_snek.http_client import HTTPClient
    from dis_snek.models.snowflake import Snowflake_Type


class EmojiRequests:
    request: "HTTPClient.request"

    async def get_all_guild_emoji(self, guild_id: "Snowflake_Type") -> List[dict]:
        """
        Get all the emoji from a guild.

        parameters:
            guild_id: The ID of the guild to query.

        Returns:
            List of emoji objects
        """
        return await self.request(Route("GET", f"/guilds/{guild_id}/emojis"))

    async def get_guild_emoji(self, guild_id: "Snowflake_Type", emoji_id: "Snowflake_Type") -> dict:
        """
        Get a specific guild emoji object

        parameters:
            guild_id: The ID of the guild to query
            emoji_id: The ID of the emoji to get

        Returns:
            Emoji object
        """
        return await self.request(Route("GET", f"/guilds/{guild_id}/emojis/{emoji_id}"))

    async def create_guild_emoji(self, payload: dict, guild_id: "Snowflake_Type", reason: str = MISSING) -> dict:
        """
        Create a guild emoji.

        parameters:
            payload: The emoji's data
            guild_id: The ID of the guild
            reason: The reason for this creation

        Returns:
            The created emoji object
        """
        return await self.request(Route("POST", f"/guilds/{guild_id}/emojis"), data=payload, reason=reason)

    async def modify_guild_emoji(
        self, payload: dict, guild_id: "Snowflake_Type", emoji_id: "Snowflake_Type", reason: str = MISSING
    ) -> dict:
        """
        Modify an existing guild emoji.

        parameters:
            payload: The emoji's data
            guild_id: The ID of the guild
            emoji_id: The ID of the emoji to update
            reason: The reason for this creation

        Returns:
            The updated emoji object
        """
        return await self.request(Route("PATCH", f"/guilds/{guild_id}/emojis/{emoji_id}"), data=payload, reason=reason)

    async def delete_guild_emoji(
        self, guild_id: "Snowflake_Type", emoji_id: "Snowflake_Type", reason: str = MISSING
    ) -> None:
        """
        Delete a guild emoji.

        Args:
            guild_id: The ID of the guild
            emoji_id: The ID of the emoji to update
            reason: The reason for this deletion
        """
        await self.request(Route("DELETE", f"/guilds/{guild_id}/emojis/{emoji_id}"), reason=reason)



from io import IOBase
from typing import TYPE_CHECKING, List, Optional, Union

import attr
from aiohttp import FormData
from attr.converters import optional

from dis_snek.const import MISSING, PREMIUM_GUILD_LIMITS
from dis_snek.models.color import Color
from dis_snek.models.discord import DiscordObject
from dis_snek.models.discord_objects.channel import (
    GuildText,
    GuildVoice,
    GuildStageVoice,
    PermissionOverwrite,
    GuildChannel,
)
from dis_snek.models.discord_objects.emoji import CustomEmoji
from dis_snek.models.discord_objects.sticker import Sticker
from dis_snek.models.discord_objects.thread import ThreadList
from dis_snek.models.enums import (
    NSFWLevels,
    Permissions,
    SystemChannelFlags,
    VerificationLevels,
    DefaultNotificationLevels,
    ExplicitContentFilterLevels,
    MFALevels,
    ChannelTypes,
)
from dis_snek.models.snowflake import to_snowflake
from dis_snek.utils.attr_utils import define
from dis_snek.utils.converters import timestamp_converter
from dis_snek.utils.serializer import to_image_data, dict_filter_none

if TYPE_CHECKING:
    from pathlib import Path

    from dis_snek.models.discord_objects.channel import TYPE_GUILD_CHANNEL, ThreadChannel, GuildCategory
    from dis_snek.models.discord_objects.role import Role
    from dis_snek.models.discord_objects.user import Member, User
    from dis_snek.models.snowflake import Snowflake_Type


@define()
class GuildBan:
    reason: Optional[str]
    user: "User"


@define()
class Guild(DiscordObject):
    """Guilds in Discord represent an isolated collection of users and channels, and are often referred to as "servers" in the UI."""

    unavailable: bool = attr.ib(default=False)
    """True if this guild is unavailable due to an outage."""
    name: str = attr.ib()
    """Name of guild. (2-100 characters, excluding trailing and leading whitespace)"""
    splash: Optional[str] = attr.ib(default=None)
    """Hash for splash image."""
    discovery_splash: Optional[str] = attr.ib(default=None)
    """Hash for discovery splash image. Only present for guilds with the "DISCOVERABLE" feature."""
    # owner: bool = attr.ib(default=False)  # we get this from api but it's kinda useless to store
    permissions: Optional[str] = attr.ib(default=None)  # todo convert to permissions obj
    """Total permissions for the user in the guild. (excludes overwrites)"""
    afk_channel_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """The channel id for afk."""
    afk_timeout: Optional[int] = attr.ib(default=None)
    """afk timeout in seconds."""
    widget_enabled: bool = attr.ib(default=False)
    """True if the server widget is enabled."""
    widget_channel_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """The channel id that the widget will generate an invite to, or None if set to no invite."""
    verification_level: Union[VerificationLevels, int] = attr.ib(default=VerificationLevels.NONE)
    """The verification level required for the guild."""
    default_message_notifications: Union[DefaultNotificationLevels, int] = attr.ib(
        default=DefaultNotificationLevels.ALL_MESSAGES
    )
    """The default message notifications level."""
    explicit_content_filter: Union[ExplicitContentFilterLevels, int] = attr.ib(
        default=ExplicitContentFilterLevels.DISABLED
    )
    """The explicit content filter level."""
    mfa_level: Union[MFALevels, int] = attr.ib(default=MFALevels.NONE)
    """The required MFA (Multi Factor Authentication) level for the guild."""
    system_channel_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """The id of the channel where guild notices such as welcome messages and boost events are posted."""
    system_channel_flags: Union[SystemChannelFlags, int] = attr.ib(default=SystemChannelFlags.NONE)
    """The system channel flags."""
    rules_channel_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """The id of the channel where Community guilds can display rules and/or guidelines."""
    joined_at: str = attr.ib(default=None, converter=optional(timestamp_converter))
    """When this guild was joined at."""
    large: bool = attr.ib(default=False)
    """True if this is considered a large guild."""
    member_count: int = attr.ib(default=0)
    """The total number of members in this guild."""
    voice_states: List[dict] = attr.ib(factory=list)
    """The states of members currently in voice channels. Lacks the guild_id key."""
    presences: List[dict] = attr.ib(factory=list)
    """The presences of the members in the guild, will only include non-offline members if the size is greater than large threshold."""
    max_presences: Optional[int] = attr.ib(default=None)
    """The maximum number of presences for the guild. (None is always returned, apart from the largest of guilds)"""
    max_members: Optional[int] = attr.ib(default=None)
    """The maximum number of members for the guild."""
    vanity_url_code: Optional[str] = attr.ib(default=None)
    """The vanity url code for the guild."""
    description: Optional[str] = attr.ib(default=None)
    """The description of a Community guild."""
    banner: Optional[str] = attr.ib(default=None)
    """Hash for banner image."""
    premium_tier: Optional[str] = attr.ib(default=None)
    """The premium tier level. (Server Boost level)"""
    premium_subscription_count: int = attr.ib(default=0)
    """The number of boosts this guild currently has."""
    preferred_locale: str = attr.ib()
    """The preferred locale of a Community guild. Used in server discovery and notices from Discord. Defaults to \"en-US\""""
    public_updates_channel_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """The id of the channel where admins and moderators of Community guilds receive notices from Discord."""
    max_video_channel_users: int = attr.ib(default=0)
    """The maximum amount of users in a video channel."""
    welcome_screen: Optional[dict] = attr.ib(factory=list)  # TODO 	welcome screen object.
    """The welcome screen of a Community guild, shown to new members, returned in an Invite's guild object."""
    nsfw_level: Union[NSFWLevels, int] = attr.ib(default=NSFWLevels.DEFAULT)
    """The guild NSFW level."""
    stage_instances: List[dict] = attr.ib(factory=list)  # TODO stage instance objects
    """Stage instances in the guild."""

    _owner_id: "Snowflake_Type" = attr.ib()
    _channel_ids: List["Snowflake_Type"] = attr.ib(factory=list)
    _thread_ids: List["Snowflake_Type"] = attr.ib(factory=list)
    _member_ids: List["Snowflake_Type"] = attr.ib(factory=list)
    _role_ids: List["Snowflake_Type"] = attr.ib(factory=list)
    _features: List[str] = attr.ib(factory=list)
    _icon: Optional[str] = attr.ib(default=None)  # todo merge, convert to asset
    _icon_hash: Optional[str] = attr.ib(default=None)

    # TODO Not storing these for now, get accurate data from api when needed instead.
    # _emojis: List[dict] = attr.ib(factory=list)
    # _stickers: List[Sticker] = attr.ib(factory=list)

    @classmethod
    def _process_dict(cls, data, client):
        guild_id = data["id"]

        channels_data = data.pop("channels", [])
        for c in channels_data:
            c["guild_id"] = guild_id
        data["channel_ids"] = [client.cache.place_channel_data(channel_data).id for channel_data in channels_data]

        threads_data = data.pop("threads", [])
        data["thread_ids"] = [client.cache.place_channel_data(thread_data).id for thread_data in threads_data]

        members_data = data.pop("members", [])
        data["member_ids"] = [client.cache.place_member_data(guild_id, member_data).id for member_data in members_data]

        roles_data = data.pop("roles", [])
        data["role_ids"] = list(client.cache.place_role_data(guild_id, roles_data).keys())

        return data

    @property
    def channels(self) -> List["TYPE_GUILD_CHANNEL"]:
        """Returns a list of channels associated with this guild."""
        return [self._client.cache.channel_cache.get(c_id) for c_id in self._channel_ids]

    @property
    def threads(self) -> List["ThreadChannel"]:
        """Returns a list of threads associated with this guild."""
        return [self._client.cache.channel_cache.get(t_id) for t_id in self._thread_ids]

    @property
    def members(self) -> List["Member"]:
        """A generator that yields all members of this guild."""
        return [self._client.cache.member_cache.get(m_id) for m_id in self._member_ids]

    @property
    def roles(self) -> List["Role"]:
        """Returns a list of roles associated with this guild"""
        return [self._client.cache.role_cache.get(r_id) for r_id in self._role_ids]

    @property
    def me(self) -> "Member":
        """Returns this bots member object within this guild."""
        return self._client.cache.member_cache.get((self.id, self._client.user.id))

    @property
    def system_channel(self) -> Optional["GuildText"]:
        """Returns the channel this guild uses for system messages."""
        return self._client.cache.channel_cache.get(self.system_channel_id)

    @property
    def rules_channel(self) -> Optional["GuildText"]:
        """Returns the channel declared as a rules channel"""
        return self._client.cache.channel_cache.get(self.rules_channel_id)

    @property
    def public_updates_channel(self) -> Optional["GuildText"]:
        """Returns the channel where server staff receive notices from Discord"""
        return self._client.cache.channel_cache.get(self.public_updates_channel_id)

    @property
    def emoji_limit(self) -> int:
        """The maximum number of emoji this guild can have"""
        base = 200 if "MORE_EMOJI" in self._features else 50
        return max(base, PREMIUM_GUILD_LIMITS[self.premium_tier]["emoji"])

    @property
    def sticker_limit(self) -> int:
        """The maximum number of stickers this guild can have"""
        base = 60 if "MORE_sTICKERS" in self._features else 0
        return max(base, PREMIUM_GUILD_LIMITS[self.premium_tier]["stickers"])

    @property
    def bitrate_limit(self) -> int:
        """The maximum bitrate for this guild"""
        base = 128000 if "VIP_REGIONS" in self._features else 96000
        return max(base, PREMIUM_GUILD_LIMITS[self.premium_tier]["bitrate"])

    @property
    def filesize_limit(self) -> int:
        """The maximum filesize that may be uploaded within this guild"""
        return PREMIUM_GUILD_LIMITS[self.premium_tier]["filesize"]

    @property
    def default_role(self) -> "Role":
        """The `@everyone` role in this guild"""
        return self._client.cache.role_cache.get(self.id)  # type: ignore

    @property
    def premium_subscriber_role(self) -> Optional["Role"]:
        """The role given to boosters of this server, if set"""
        for role in self.roles:
            if role.premium_subscriber:
                return role
        return None

    @property
    def my_role(self) -> Optional["Role"]:
        """The role associated with this client, if set"""
        m_r_id = self._client.user.id
        for role in self.roles:
            if role._bot_id == m_r_id:
                return role
        return None

    async def get_owner(self) -> "Member":
        # maybe precache owner instead of using `get_owner`
        return await self._client.cache.get_member(self.id, self._owner_id)

    def is_owner(self, member: "Member") -> bool:
        return self._owner_id == member.id

    async def edit(
        self,
        name: Optional[str] = MISSING,
        description: Optional[str] = MISSING,
        verification_level: Optional["VerificationLevels"] = MISSING,
        default_message_notifications: Optional["DefaultNotificationLevels"] = MISSING,
        explicit_content_filter: Optional["ExplicitContentFilterLevels"] = MISSING,
        afk_channel: Optional[Union["GuildVoice", "Snowflake_Type"]] = MISSING,
        afk_timeout: Optional[int] = MISSING,
        system_channel: Optional[Union["GuildText", "Snowflake_Type"]] = MISSING,
        system_channel_flags: Optional[SystemChannelFlags] = MISSING,
        # ToDo: these are not tested. Mostly, since I do not have access to those features
        owner: Optional[Union["Member", "Snowflake_Type"]] = MISSING,
        icon: Optional[Union[str, "Path", "IOBase"]] = MISSING,
        splash: Optional[Union[str, "Path", "IOBase"]] = MISSING,
        discovery_splash: Optional[Union[str, "Path", "IOBase"]] = MISSING,
        banner: Optional[Union[str, "Path", "IOBase"]] = MISSING,
        rules_channel: Optional[Union["GuildText", "Snowflake_Type"]] = MISSING,
        public_updates_channel: Optional[Union["GuildText", "Snowflake_Type"]] = MISSING,
        preferred_locale: Optional[str] = MISSING,
        # ToDo: validate voice region
        region: Optional[str] = MISSING,
        # ToDo: Fill in guild features. No idea how this works - https://discord.com/developers/docs/resources/guild#guild-object-guild-features
        features: Optional[list[str]] = MISSING,
        reason: Optional[str] = MISSING,
    ):
        """
        Edit the guild.

        Parameters:
            name: The new name of the guild.
            description: The new description of the guild.
            region: ToDo
            verification_level: The new verification level for the guild.
            default_message_notifications: The new notification level for the guild.
            explicit_content_filter: The new explicit content filter level for the guild.
            afk_channel: The voice channel that should be the new AFK channel.
            afk_timeout: How many seconds does a member need to be afk before they get moved to the AFK channel. Must be either `60`, `300`, `900`, `1800` or `3600`, otherwise HTTPException will be raised.
            icon: The new icon. Requires a bytes like object or a path to an image.
            owner: The new owner of the guild. You, the bot, need to be owner for this to work.
            splash: The new invite splash image. Requires a bytes like object or a path to an image.
            discovery_splash: The new discovery image. Requires a bytes like object or a path to an image.
            banner: The new banner image. Requires a bytes like object or a path to an image.
            system_channel: The text channel where new system messages should appear. This includes boosts and welcome messages.
            system_channel_flags: The new settings for the system channel.
            rules_channel: The text channel where your rules and community guidelines are displayed.
            public_updates_channel: The text channel where updates from discord should appear.
            preferred_locale: The new preferred locale of the guild. Must be an ISO 639 code.
            features: ToDo
            reason: An optional reason for the audit log.
        """

        await self._client.http.modify_guild(
            guild_id=self.id,
            name=name,
            description=description,
            region=region,
            verification_level=int(verification_level) if verification_level else MISSING,
            default_message_notifications=int(default_message_notifications)
            if default_message_notifications
            else MISSING,
            explicit_content_filter=int(explicit_content_filter) if explicit_content_filter else MISSING,
            afk_channel_id=to_snowflake(afk_channel) if afk_channel else MISSING,
            afk_timeout=afk_timeout,
            icon=to_image_data(icon) if icon else MISSING,
            owner_id=to_snowflake(owner) if owner else MISSING,
            splash=to_image_data(splash) if splash else MISSING,
            discovery_splash=to_image_data(discovery_splash) if discovery_splash else MISSING,
            banner=to_image_data(banner) if banner else MISSING,
            system_channel_id=to_snowflake(system_channel) if system_channel else MISSING,
            system_channel_flags=int(system_channel_flags) if system_channel_flags else MISSING,
            rules_channel_id=to_snowflake(rules_channel) if rules_channel else MISSING,
            public_updates_channel_id=to_snowflake(public_updates_channel) if public_updates_channel else MISSING,
            preferred_locale=preferred_locale,
            features=features,
            reason=reason,
        )

    async def create_custom_emoji(
        self,
        name: str,
        imagefile: Union[str, "Path", "IOBase"],
        roles: Optional[List[Union["Snowflake_Type", "Role"]]] = None,
        reason: Optional[str] = MISSING,
    ) -> "CustomEmoji":
        """
        Create a new custom emoji for the guild.

        parameters:
            name: Name of the emoji
            imagefile: The emoji image. (Supports PNG, JPEG, WebP, GIF)
            roles: Roles allowed to use this emoji.
            reason: An optional reason for the audit log.

        returns:
            The new custom emoji created.
        """
        data_payload = dict_filter_none(
            dict(
                name=name,
                image=to_image_data(imagefile),
                roles=roles,
            )
        )

        emoji_data = await self._client.http.create_guild_emoji(data_payload, self.id, reason=reason)
        emoji_data["guild_id"] = self.id
        return CustomEmoji.from_dict(emoji_data, self._client)  # TODO Probably cache it

    async def get_all_custom_emojis(self) -> List[CustomEmoji]:
        """
        Gets all the custom emoji present for this guild.

        returns:
            A list of custom emoji objects.
        """
        emojis_data = await self._client.http.get_all_guild_emoji(self.id)
        return [CustomEmoji.from_dict(emoji_data, self._client) for emoji_data in emojis_data]

    async def get_custom_emoji(self, emoji_id: "Snowflake_Type") -> CustomEmoji:
        """
        Gets the custom emoji present for this guild, based on the emoji id.

        parameters:
            emoji_id: The target emoji to get data of.

        returns:
            The custom emoji object.
        """
        emoji_data = await self._client.http.get_guild_emoji(self.id, emoji_id)
        return CustomEmoji.from_dict(emoji_data, self._client)

    async def create_channel(
        self,
        channel_type: Union[ChannelTypes, int],
        name: str,
        topic: Optional[str] = MISSING,
        position: int = 0,
        permission_overwrites: Optional[List[Union["PermissionOverwrite", dict]]] = MISSING,
        category: Union["Snowflake_Type", "GuildCategory"] = None,
        nsfw: bool = False,
        bitrate: int = 64000,
        user_limit: int = 0,
        slowmode_delay: int = 0,
        reason: Optional[str] = MISSING,
    ) -> "TYPE_GUILD_CHANNEL":
        """
        Create a guild channel, allows for explicit channel type setting.

        parameters:
            channel_type: The type of channel to create
            name: The name of the channel
            topic: The topic of the channel
            position: The position of the channel in the channel list
            permission_overwrites: Permission overwrites to apply to the channel
            category: The category this channel should be within
            nsfw: Should this channel be marked nsfw
            bitrate: The bitrate of this channel, only for voice
            user_limit: The max users that can be in this channel, only for voice
            slowmode_delay: The time users must wait between sending messages
            reason: The reason for creating this channel

        returns:
            The newly created channel.
        """
        if category:
            category = to_snowflake(category)

        channel_data = await self._client.http.create_guild_channel(
            self.id,
            name,
            channel_type,
            topic,
            position,
            permission_overwrites,
            category,
            nsfw,
            bitrate,
            user_limit,
            slowmode_delay,
            reason,
        )
        return self._client.cache.place_channel_data(channel_data)

    async def create_text_channel(
        self,
        name: str,
        topic: Optional[str] = MISSING,
        position: int = 0,
        permission_overwrites: Optional[List[Union["PermissionOverwrite", dict]]] = MISSING,
        category: Union["Snowflake_Type", "GuildCategory"] = None,
        nsfw: bool = False,
        slowmode_delay: int = 0,
        reason: Optional[str] = MISSING,
    ) -> "GuildText":
        """
        Create a text channel in this guild.

        parameters:
            name: The name of the channel
            topic: The topic of the channel
            position: The position of the channel in the channel list
            permission_overwrites: Permission overwrites to apply to the channel
            category: The category this channel should be within
            nsfw: Should this channel be marked nsfw
            slowmode_delay: The time users must wait between sending messages
            reason: The reason for creating this channel

        returns:
           The newly created text channel.
        """
        return await self.create_channel(
            channel_type=ChannelTypes.GUILD_TEXT,
            name=name,
            topic=topic,
            position=position,
            permission_overwrites=permission_overwrites,
            category=category,
            nsfw=nsfw,
            slowmode_delay=slowmode_delay,
            reason=reason,
        )

    async def create_voice_channel(
        self,
        name: str,
        topic: Optional[str] = MISSING,
        position: int = 0,
        permission_overwrites: Optional[List[Union["PermissionOverwrite", dict]]] = MISSING,
        category: Union["Snowflake_Type", "GuildCategory"] = None,
        nsfw: bool = False,
        bitrate: int = 64000,
        user_limit: int = 0,
        reason: Optional[str] = MISSING,
    ) -> "GuildVoice":
        """
        Create a guild voice channel.

        parameters:
            name: The name of the channel
            topic: The topic of the channel
            position: The position of the channel in the channel list
            permission_overwrites: Permission overwrites to apply to the channel
            category: The category this channel should be within
            nsfw: Should this channel be marked nsfw
            bitrate: The bitrate of this channel, only for voice
            user_limit: The max users that can be in this channel, only for voice
            reason: The reason for creating this channel

        returns:
           The newly created voice channel.
        """
        return await self.create_channel(
            channel_type=ChannelTypes.GUILD_VOICE,
            name=name,
            topic=topic,
            position=position,
            permission_overwrites=permission_overwrites,
            category=category,
            nsfw=nsfw,
            bitrate=bitrate,
            user_limit=user_limit,
            reason=reason,
        )

    async def create_stage_channel(
        self,
        name: str,
        topic: Optional[str] = MISSING,
        position: int = 0,
        permission_overwrites: Optional[List[Union["PermissionOverwrite", dict]]] = MISSING,
        category: Union["Snowflake_Type", "GuildCategory"] = MISSING,
        bitrate: int = 64000,
        user_limit: int = 0,
        reason: Optional[str] = MISSING,
    ) -> "GuildStageVoice":
        """
        Create a guild stage channel.

        parameters:
            name: The name of the channel
            topic: The topic of the channel
            position: The position of the channel in the channel list
            permission_overwrites: Permission overwrites to apply to the channel
            category: The category this channel should be within
            bitrate: The bitrate of this channel, only for voice
            user_limit: The max users that can be in this channel, only for voice
            reason: The reason for creating this channel

        returns:
            The newly created stage channel.
        """
        return await self.create_channel(
            channel_type=ChannelTypes.GUILD_STAGE_VOICE,
            name=name,
            topic=topic,
            position=position,
            permission_overwrites=permission_overwrites,
            category=category,
            bitrate=bitrate,
            user_limit=user_limit,
            reason=reason,
        )

    async def create_category(
        self,
        name: str,
        position: int = 0,
        permission_overwrites: Optional[List[Union["PermissionOverwrite", dict]]] = MISSING,
        reason: Optional[str] = MISSING,
    ) -> "GuildCategory":
        """
        Create a category within this guild.

        parameters:
            name: The name of the channel
            position: The position of the channel in the channel list
            permission_overwrites: Permission overwrites to apply to the channel
            reason: The reason for creating this channel

        returns:
            The newly created category.
        """
        return await self.create_channel(
            channel_type=ChannelTypes.GUILD_CATEGORY,
            name=name,
            position=position,
            permission_overwrites=permission_overwrites,
            reason=reason,
        )

    async def delete_channel(self, channel: Union["TYPE_GUILD_CHANNEL", "Snowflake_Type"], reason: str = None) -> None:
        """
        Delete the given channel, can handle either a snowflake or channel object

        This is effectively just an alias for `channel.delete()`

        Args:
            channel: The channel to be deleted
            reason: The reason for this deletion
        """
        if isinstance(channel, (str, int)):
            channel = await self._client.get_channel(channel)

        if not channel:
            raise ValueError("Unable to find requested channel")

        if channel.id not in self._channel_ids:
            raise ValueError("This guild does not hold the requested channel")

        await channel.delete(reason)

    async def create_custom_sticker(
        self,
        name: str,
        imagefile: Union[str, "Path", "IOBase"],
        description: Optional[str] = MISSING,
        tags: Optional[str] = MISSING,
        reason: Optional[str] = MISSING,
    ) -> "Sticker":
        """
        # TODO
        """
        payload = FormData()
        payload.add_field("name", name)

        # TODO Validate image type?
        if isinstance(imagefile, IOBase):
            payload.add_field("file", name)
        else:
            payload.add_field("file", open(str(imagefile)))

        if description:
            payload.add_field("description", description)

        if tags:
            payload.add_field("tags", tags)

        sticker_data = await self._client.http.create_guild_sticker(payload, self.id, reason)
        return Sticker.from_dict(sticker_data, self._client)

    async def get_all_custom_stickers(self) -> List["Sticker"]:
        """
        # TODO
        """
        stickers_data = await self._client.http.list_guild_stickers(self.id)
        return Sticker.from_list(stickers_data, self._client)

    async def get_custom_sticker(self, sticker_id: "Snowflake_Type") -> "Sticker":
        """
        # TODO
        """
        sticker_data = await self._client.http.get_guild_sticker(self.id, to_snowflake(sticker_id))
        return Sticker.from_dict(sticker_data, self._client)

    async def get_active_threads(self) -> "ThreadList":
        """
        Gets all active threads in the guild, including public and private threads. Threads are ordered by their id, in descending order.

        returns:
            List of active threads and thread member object for each returned thread the bot user has joined.
        """
        threads_data = await self._client.http.list_active_threads(self.id)
        return ThreadList.from_dict(threads_data, self._client)

    async def get_role(self, role_id: "Snowflake_Type") -> Optional["Role"]:
        """
        Get the specified role by ID.

        Args:
            role_id: The ID of the role to get

        Returns:
            A role object or None if the role is not found.
        """
        return await self._client.cache.get_role(self.id, role_id)

    async def create_role(
        self,
        name: Optional[str] = MISSING,
        permissions: Optional[Permissions] = MISSING,
        colour: Optional[Union[Color, int]] = MISSING,
        color: Optional[Union[Color, int]] = MISSING,
        hoist: Optional[bool] = False,
        mentionable: Optional[bool] = False,
        # ToDo: icon needs testing. I have to access to that
        icon: Optional[Union[str, "Path", "IOBase"]] = MISSING,
        reason: Optional[str] = MISSING,
    ) -> "Role":
        """
        Create a new role for the guild.
        You must have the `manage roles` permission.

        Args:
            name: The name the role should have. `Default: new role`
            permissions: The permissions the role should have. `Default: @everyone permissions`
            colour: The colour of the role. Can be either `Color` or an RGB integer. `Default: BrandColors.BLACK`
            color: Alias for `colour`
            icon: Can be either a bytes like object or a path to an image, or a unicode emoji which is supported by discord.
            hoist: Whether the role is shown separately in the members list. `Default: False`
            mentionable: Whether the role can be mentioned. `Default: False`
            reason: An optional reason for the audit log.

        Returns:
            A role object or None if the role is not found.
        """

        payload = {}

        if name:
            payload.update({"name": name})

        if permissions:
            payload.update({"permissions": str(int(permissions))})

        colour = colour or color
        if colour:
            payload.update({"color": colour.value})

        if hoist:
            payload.update({"hoist": True})

        if mentionable:
            payload.update({"mentionable": True})

        if icon:
            # test if the icon is probably a unicode emoji (str and len() == 1) or a path / bytes obj
            if isinstance(icon, str) and len(icon) == 1:
                payload.update({"unicode_emoji": icon})

            else:
                payload.update({"icon": to_image_data(icon)})

        result = await self._client.http.create_guild_role(guild_id=self.id, payload=payload, reason=reason)
        return self._client.cache.place_role_data(guild_id=self.id, data=[result])[to_snowflake(result["id"])]

    async def get_channel(self, channel_id: "Snowflake_Type") -> Optional[Union["GuildChannel", "ThreadChannel"]]:
        """
        Returns a channel with the given `channel_id`

        Args:
            channel_id: The ID of the channel to get

        Returns:
            Channel object if found, otherwise None
        """
        if channel_id in self._channel_ids and channel_id not in self._thread_ids:
            # theoretically, this could get any channel the client can see,
            # but to make it less confusing to new programmers,
            # i intentionally check that the guild contains the channel first
            return await self._client.cache.get_channel(channel_id)
        return None

    async def get_thread(self, thread_id: "Snowflake_Type") -> Optional["ThreadChannel"]:
        """
        Returns a Thread with the given `thread_id`

        Args:
            thread_id: The ID of the thread to get

        Returns:
            Channel object if found, otherwise None
        """
        # get_channel can retrieve threads, so this is basically an alias with extra steps for that
        if thread_id in self._thread_ids:
            return await self.get_channel(thread_id)
        return None

    async def prune_members(
        self,
        days: int = 7,
        roles: List[Union["Snowflake_Type", "Role"]] = MISSING,
        compute_prune_count: bool = True,
        reason: str = MISSING,
    ) -> Optional[int]:
        """
        Begin a guild prune. Removes members from the guild who who have not interacted for the last `days` days.
        By default, members with roles are excluded from pruning, to include them, pass their role (or role id) in `roles`
        Requires `kick members` permission.

        Args:
            days: number of days to prune (1-30)
            roles: list of roles to include in the prune
            compute_prune_count: Whether the number of members pruned should be calculated (disable this for large guilds)
            reason: The reason for this prune

        Returns:
            The total number of members pruned, if `compute_prune_count` is set to True, otherwise None
        """
        if roles is not MISSING:
            roles = [r.id if isinstance(r, Role) else r for r in roles]
        else:
            roles = []

        resp = await self._client.http.begin_guild_prune(
            self.id, days, include_roles=roles, compute_prune_count=compute_prune_count, reason=reason
        )
        return resp["pruned"]

    async def estimate_prune_members(
        self, days: int = 7, roles: List[Union["Snowflake_Type", "Role"]] = MISSING
    ) -> int:
        """
        Calculate how many members would be pruned, should `guild.prune_members` be used.
        By default, members with roles are excluded from pruning, to include them, pass their role (or role id) in `roles`

        Args:
            days: number of days to prune (1-30)
            roles: list of roles to include in the prune

        Returns:
            Total number of members that would be pruned
        """

        if roles is not MISSING:
            roles = [r.id if isinstance(r, Role) else r for r in roles]
        else:
            roles = []

        resp = await self._client.http.get_guild_prune_count(self.id, days=days, include_roles=roles)
        return resp["pruned"]

    async def leave(self) -> None:
        """Leave this guild"""
        await self._client.http.leave_guild(self.id)

    async def delete(self) -> None:
        """Delete the guild. You must own this guild to do this."""
        await self._client.http.delete_guild(self.id)

    async def kick(self, user: Union["User", "Member", "Snowflake_Type"], reason: str = MISSING) -> None:
        """
        Kick a user from the guild.
        You must have the `kick members` permission
        Args:
            user: The user to kick
            reason: The reason for the kick
        """
        await self._client.http.remove_guild_member(self.id, to_snowflake(user), reason=reason)

    async def ban(
        self, user: Union["User", "Member", "Snowflake_Type"], delete_message_days: int = 0, reason: str = MISSING
    ) -> None:
        """
        Ban a user from the guild.
        You must have the `ban members` permission
        Args:
            user: The user to ban
            delete_message_days: How many days worth of messages to remove
            reason: The reason for the ban
        """
        await self._client.http.create_guild_ban(self.id, to_snowflake(user), delete_message_days, reason=reason)

    async def get_ban(self, user: Union["User", "Member", "Snowflake_Type"]) -> GuildBan:
        """
        Get's the ban information for the specified user in the guild.
        You must have the `ban members` permission

        Args:
            user: The user to look up.

        Raises:
            NotFound: If the user is not banned in the guild.

        Returns:
            The ban information.
        """

        ban_info = await self._client.http.get_guild_ban(self.id, to_snowflake(user))
        return GuildBan(reason=ban_info["reason"], user=self._client.cache.place_user_data(ban_info["user"]))

    async def get_bans(self) -> list[GuildBan]:
        """
        Get's all bans for the guild.
        You must have the `ban members` permission

        Returns:
            A list containing all bans and information about them.
        """

        ban_infos = await self._client.http.get_guild_bans(self.id)
        return [
            GuildBan(reason=ban_info["reason"], user=self._client.cache.place_user_data(ban_info["user"]))
            for ban_info in ban_infos
        ]

    async def unban(self, user: Union["User", "Member", "Snowflake_Type"], reason: str = MISSING) -> None:
        """
        Unban a user from the guild.
        You must have the `ban members` permission
        Args:
            user: The user to unban
            reason: The reason for the ban
        """
        await self._client.http.remove_guild_ban(self.id, to_snowflake(user), reason=reason)

    async def get_widget_image(self, style: str = None) -> str:
        """
        Get a guilds widget image

        For a list of styles, look here: https://discord.com/developers/docs/resources/guild#get-guild-widget-image-widget-style-options

        Args:
            style: The style to use for the widget image
        """
        return await self._client.http.get_guild_widget_image(self.id, style)

    async def get_widget(self) -> dict:
        """
        Gets the guilds widget
        """
        # todo: Guild widget object
        return await self._client.http.get_guild_widget(self.id)

    async def modify_widget(
        self, enabled: bool = None, channel: Union["TYPE_GUILD_CHANNEL", "Snowflake_Type"] = None
    ) -> dict:
        """
        Modify the guild's widget.

        Args:
            enabled: Should the widget be enabled?
            channel: The channel to use in the widget
        """
        if channel:
            if isinstance(channel, DiscordObject):
                channel = channel.id
        return await self._client.http.modify_guild_widget(self.id, enabled, channel)



import asyncio
import re
from enum import IntEnum
from typing import TYPE_CHECKING, Callable, Coroutine, Dict, List, Union, Optional, Any

import attr

from dis_snek.const import (
    GLOBAL_SCOPE,
    CONTEXT_MENU_NAME_LENGTH,
    SLASH_OPTION_NAME_LENGTH,
    SLASH_CMD_NAME_LENGTH,
    SLASH_CMD_MAX_OPTIONS,
    SLASH_CMD_MAX_DESC_LENGTH,
)
from dis_snek.mixins.serialization import DictSerializationMixin
from dis_snek.models.command import BaseCommand
from dis_snek.models.discord_objects.channel import BaseChannel
from dis_snek.models.discord_objects.role import Role
from dis_snek.models.discord_objects.user import BaseUser
from dis_snek.models.enums import CommandTypes
from dis_snek.models.snowflake import to_snowflake
from dis_snek.utils.serializer import no_export_meta

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type
    from dis_snek.models.context import InteractionContext


class OptionTypes(IntEnum):
    """Option types supported by slash commands."""

    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10

    @classmethod
    def from_type(cls, t: type) -> "OptionTypes":
        """
        Convert data types to their corresponding OptionType.

        parameters:
            t: The datatype to convert

        returns:
            OptionType or None
        """
        if issubclass(t, str):
            return cls.STRING
        if issubclass(t, int):
            return cls.INTEGER
        if issubclass(t, bool):
            return cls.BOOLEAN
        if issubclass(t, BaseUser):
            return cls.USER
        if issubclass(t, BaseChannel):
            return cls.CHANNEL
        if issubclass(t, Role):
            return cls.ROLE
        if issubclass(t, float):
            return cls.NUMBER


class PermissionTypes(IntEnum):
    """Types of target supported by the interaction permission."""

    ROLE = 1
    USER = 2

    @classmethod
    def from_type(cls, t: type) -> "PermissionTypes":
        if issubclass(t, Role):
            return cls.ROLE
        if issubclass(t, BaseUser):
            return cls.USER


class CallbackTypes(IntEnum):
    """Types of callback supported by interaction response."""

    PONG = 1
    CHANNEL_MESSAGE_WITH_SOURCE = 4
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7


@attr.s(slots=True)
class Permission:
    """
    Represents a interaction permission.

    parameters:
        id: The id of the role or user.
        type: The type of id (user or role)
        permission: The state of permission. ``True`` to allow, ``False``, to disallow.
    """

    id: "Snowflake_Type" = attr.ib(converter=to_snowflake)
    type: Union[PermissionTypes, int] = attr.ib(converter=PermissionTypes)
    permission: bool = attr.ib(default=True)

    def to_dict(self) -> dict:
        """
        Convert this object into a dict ready for discord.

        returns:
            Representation of this object
        """
        return attr.asdict(self)


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class InteractionCommand(BaseCommand):
    """
    Represents a discord abstract interaction command.

    parameters:
        scope: Denotes whether its global or for specific guild.
        default_permission: Is this command available to all users?
        permissions: Map of guild id and its respective list of permissions to apply.
        cmd_id: The id of this command given by discord.
        callback: The coroutine to callback when this interaction is received.
    """

    name: str = attr.ib()
    """1-32 character name"""

    scope: "Snowflake_Type" = attr.ib(default=GLOBAL_SCOPE, converter=to_snowflake, metadata=no_export_meta)
    """The scope of this interaction. Global or guild ids"""
    default_permission: bool = attr.ib(default=True)
    """whether the command is enabled by default when the app is added to a guild"""
    permissions: Dict["Snowflake_Type", Union[Permission, Dict]] = attr.ib(factory=dict)
    """The permissions of this interaction"""

    cmd_id: "Snowflake_Type" = attr.ib(default=None, metadata=no_export_meta)
    """The unique ID of this interaction"""
    callback: Callable[..., Coroutine] = attr.ib(default=None, metadata=no_export_meta)
    """The coroutine to call when this interaction is received"""

    @property
    def resolved_name(self):
        """A representation of this interaction's name"""
        return self.name


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class ContextMenu(InteractionCommand):
    """
    Represents a discord context menu.

    parameters:
        name: The name of this entry.
        type: The type of entry (user or message).
    """

    name: str = attr.ib()
    """1-32 character name"""
    type: CommandTypes = attr.ib()
    """the type of command, defaults 1 if not set"""

    @name.validator
    def _name_validator(self, attribute: str, value: str) -> None:
        if not 1 <= len(value) <= CONTEXT_MENU_NAME_LENGTH:
            raise ValueError("Context Menu name attribute must be between 1 and 32 characters")

    @type.validator
    def _type_validator(self, attribute: str, value: int):
        if not isinstance(value, CommandTypes):
            if value not in CommandTypes.__members__.values():
                raise ValueError("Context Menu type not recognised, please consult the docs.")
        elif value == CommandTypes.CHAT_INPUT:
            raise ValueError(
                "The CHAT_INPUT type is basically slash commands. Please use the @slash_command() " "decorator instead."
            )


@attr.s(slots=True)
class SlashCommandChoice(DictSerializationMixin):
    """
    Represents a discord slash command choice.

    parameters:
        name: The name the user will see
        value: The data sent to your code when this choice is used
    """

    name: str = attr.ib()
    value: Union[str, int, float] = attr.ib()


@attr.s(slots=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class SlashCommandOption(DictSerializationMixin):
    """
    Represents a discord slash command option.

    parameters:
        name: The name of this option
        type: The type of option
        description: The description of this option
        required: "This option must be filled to use the command"
        choices: A list of choices the user has to pick between
    """

    name: str = attr.ib()
    type: Union[OptionTypes, int] = attr.ib()
    description: str = attr.ib(default="No Description Set")
    required: bool = attr.ib(default=True)
    choices: List[Union[SlashCommandChoice, Dict]] = attr.ib(factory=list)

    @name.validator
    def _name_validator(self, attribute: str, value: str) -> None:
        if not re.match(rf"^[\w-]{{1,{SLASH_CMD_NAME_LENGTH}}}$", value) or value != value.lower():
            raise ValueError(
                f"Options names must be lower case and match this regex: ^[\w-]{1,{SLASH_CMD_NAME_LENGTH}}$"
            )  # noqa: W605

    @description.validator
    def _description_validator(self, attribute: str, value: str) -> None:
        if not 1 <= len(value) <= SLASH_OPTION_NAME_LENGTH:
            raise ValueError("Options must be between 1 and 100 characters long")

    @type.validator
    def _type_validator(self, attribute: str, value: int) -> None:
        if value == OptionTypes.SUB_COMMAND or value == OptionTypes.SUB_COMMAND_GROUP:
            raise ValueError(
                "Options cannot be SUB_COMMAND or SUB_COMMAND_GROUP. If you want to use subcommands, "
                "see the @sub_command() decorator."
            )


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class SlashCommand(InteractionCommand):
    """
    Represents a discord slash command.

    parameters:
        name: The name of this command.
        description: The description of this command.
        options: A list of options for this command.
    """

    name: str = attr.ib()
    description: str = attr.ib(default="No Description Set")
    options: List[Union[SlashCommandOption, Dict]] = attr.ib(factory=list)
    subcommand_callbacks: dict = attr.ib(factory=dict, metadata=no_export_meta)

    def __attrs_post_init__(self):
        if self.callback is not None:
            if hasattr(self.callback, "options"):
                if not self.options:
                    self.options = []
                self.options += self.callback.options

            if hasattr(self.callback, "permissions"):
                self.permissions = self.callback.permissions
        super().__attrs_post_init__()

    @name.validator
    def _name_validator(self, attribute: str, value: str) -> None:
        if not re.match(rf"^[\w-]{{1,{SLASH_CMD_NAME_LENGTH}}}$", value) or value != value.lower():
            raise ValueError(
                f"Slash Command names must be lower case and match this regex: ^[\w-]{1, {SLASH_CMD_NAME_LENGTH} }$"
            )  # noqa: W605

    @description.validator
    def _description_validator(self, attribute: str, value: str) -> None:
        if not 1 <= len(value) <= SLASH_CMD_MAX_DESC_LENGTH:
            raise ValueError(f"Description must be between 1 and {SLASH_CMD_MAX_DESC_LENGTH} characters long")

    @options.validator
    def _options_validator(self, attribute: str, value: List) -> None:
        if value:
            if isinstance(value, list):
                if len(value) > SLASH_CMD_MAX_OPTIONS:
                    raise ValueError(f"Slash commands can only hold {SLASH_CMD_MAX_OPTIONS} options")
            else:
                raise TypeError("Options attribute must be either None or a list of options")

            if any(opt.type in (OptionTypes.SUB_COMMAND, OptionTypes.SUB_COMMAND_GROUP) for opt in value):
                if any(opt.type not in (OptionTypes.SUB_COMMAND, OptionTypes.SUB_COMMAND_GROUP) for opt in value):
                    raise ValueError("Options aren't supported when subcommands are defined")


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class SubCommand(SlashCommand):
    options: List[Union[SlashCommandOption, Dict]] = attr.ib(factory=list)

    group_name: str = attr.ib(default=None, metadata=no_export_meta)
    group_description: str = attr.ib(default=None, metadata=no_export_meta)

    subcommand_name: str = attr.ib(default=None, metadata=no_export_meta)
    subcommand_description: str = attr.ib(default=None, metadata=no_export_meta)

    @property
    def resolved_name(self):
        return f"{self.name} {f'{self.group_name} ' if self.group_name else ''}{self.subcommand_name}"

    async def subcommand_call(self, context: "InteractionContext", *args, **kwargs):
        if call := self.subcommand_callbacks.get(context.invoked_name):
            return await call(context, *args, **kwargs)
        raise ValueError(f"Error {context.invoked_name} is not a known subcommand")

    def child_to_dict(self):
        sub_cmd = {
            "name": self.subcommand_name,
            "description": self.subcommand_description,
            "type": OptionTypes.SUB_COMMAND,
            "options": [opt.to_dict() if hasattr(opt, "to_dict") else opt for opt in self.options],
        }
        if self.group_name:
            sub_cmd = {
                "name": self.group_name,
                "description": self.group_description,
                "type": OptionTypes.SUB_COMMAND_GROUP,
                "options": [sub_cmd],
            }
        return sub_cmd

    def to_dict(self):
        parent = super().to_dict()
        return parent


##############
# Decorators #
##############


def slash_command(
    name: str,
    description: str = "No description set",
    scope: "Snowflake_Type" = GLOBAL_SCOPE,
    options: Optional[List[Union[SlashCommandOption, Dict]]] = None,
    default_permission: bool = True,
    permissions: Optional[Dict["Snowflake_Type", Union[Permission, Dict]]] = None,
    sub_cmd_name: str = None,
    sub_cmd_description: str = "No description set",
    group_name: str = None,
    group_description: str = "No description set",
):
    """
    A decorator to declare a coroutine as a slash command.

    note:
        While the base and group descriptions arent visible in the discord client, currently.
        We strongly advise defining them anyway, if you're using subcommands, as Discord has said they will be visible in
        one of the future ui updates.

    parameters:
        name: 1-32 character name of the command
        description: 1-100 character description of the command
        scope: The scope this command exists within
        options: The parameters for the command, max 25
        default_permission: Whether the command is enabled by default when the app is added to a guild
        permissions: The roles or users who can use this command
        sub_cmd_name: 1-32 character name of the subcommand
        sub_cmd_description: 1-100 character description of the subcommand
        group_name: 1-32 character name of the group
        group_description: 1-100 character description of the group

    returns:
        SlashCommand Object
    """

    def wrapper(func) -> SlashCommand:
        if not asyncio.iscoroutinefunction(func):
            raise ValueError("Commands must be coroutines")

        if not sub_cmd_name:
            cmd = SlashCommand(
                name=name,
                description=description,
                scope=scope,
                callback=func,
                options=options,
                default_permission=default_permission,
                permissions=permissions,
            )
        else:
            cmd = SubCommand(
                name=name,
                description=description,
                subcommand_name=sub_cmd_name,
                subcommand_description=sub_cmd_description,
                group_name=group_name,
                group_description=group_description,
                options=options,
                callback=func,
                scope=scope,
                default_permission=default_permission,
                permissions=permissions,
            )

        func.cmd_id = f"{scope}::{name}"
        return cmd

    return wrapper


def sub_command(
    base_name: str,
    sub_name: str,
    group_name: str = None,
    base_description: str = "No Description set",
    sub_description: str = "No Description set",
    group_description: str = "No Description set",
    scope: "Snowflake_Type" = GLOBAL_SCOPE,
    options: Optional[List[Union[SlashCommandOption, Dict]]] = None,
    default_permission: bool = True,
    permissions: Optional[Dict["Snowflake_Type", Union[Permission, Dict]]] = None,
):
    """
    A decorator to declare a coroutine as a slash subcommand.

    note:
        While the base and group descriptions arent visible in the discord client, currently.
        We strongly advise defining them anyway, if you're using subcommands, as Discord has said they will be visible in
        one of the future ui updates.


    parameters:
        base_name: The name of the base command
        sub_name: The name of the subcommand
        group_name: The name of the command group (optional)
        base_description: The description of the base command
        sub_description: The description of the subcommand
        group_description: The description of the command group
        scope: The scope this command exists within
        options: The parameters for the command, max 25
        default_permission: Whether the command is enabled by default when the app is added to a guild
        permissions: The roles or users who can use this command
    """

    def wrapper(func) -> SlashCommand:
        cmd = SubCommand(
            name=base_name,
            description=base_description,
            subcommand_name=sub_name,
            subcommand_description=sub_description,
            group_name=group_name,
            group_description=group_description,
            options=options,
            callback=func,
            scope=scope,
            default_permission=default_permission,
            permissions=permissions,
        )

        func.cmd_id = f"{scope}::{base_name}"
        return cmd

    return wrapper


def context_menu(
    name: str,
    context_type: "CommandTypes",
    scope: "Snowflake_Type" = GLOBAL_SCOPE,
    default_permission: bool = True,
    permissions: Optional[Dict["Snowflake_Type", Union[Permission, Dict]]] = None,
):
    """
    A decorator to declare a coroutine as a Context Menu

    parameters:
        name: 1-32 character name of the context menu
        context_type: The type of context menu
        scope: The scope this command exists within
        default_permission: Whether the menu is enabled by default when the app is added to a guild
        permissions: The roles or users who can use this menu

    returns:
        ContextMenu object
    """

    def wrapper(func) -> ContextMenu:
        if not asyncio.iscoroutinefunction(func):
            raise ValueError("Commands must be coroutines")

        perm = permissions
        if hasattr(func, "permissions"):
            if perm:
                perm.update(func.permissions)
            else:
                perm = func.permissions

        cmd = ContextMenu(
            name=name,
            type=context_type,
            scope=scope,
            default_permission=default_permission,
            permissions=perm,
            callback=func,
        )
        return cmd

    return wrapper


def slash_option(
    name: str,
    description: str,
    opt_type: Union[OptionTypes, int],
    required: bool = False,
    choices: List[Union[SlashCommandChoice, dict]] = None,
) -> Any:
    """
    A decorator to add an option to a slash command.

    parameters:
        name: 1-32 lowercase character name matching ^[\w-]{1,32}$
        opt_type: The type of option
        description: 1-100 character description of option
        required: If the parameter is required or optional--default false
        choices: A list of choices the user has to pick between (max 25)
    """

    def wrapper(func):
        if hasattr(func, "cmd_id"):
            raise Exception("slash_option decorators must be positioned under a slash_command decorator")

        option = SlashCommandOption(
            name=name, type=opt_type, description=description, required=required, choices=choices if choices else []
        )

        if not hasattr(func, "options"):
            func.options = []
        func.options.append(option)
        return func

    return wrapper


def slash_permission(guild_id: "Snowflake_Type", permissions: List[Union[Permission, Dict]]) -> Any:
    """
    A decorator to add permissions for a guild to a slash command or context menu.

    parameters:
        guild_id: The target guild to apply the permissions.
        permissions: A list of interaction permission rights.
    """
    guild_id = to_snowflake(guild_id)

    def wrapper(func):
        if hasattr(func, "cmd_id"):
            raise Exception("slash_option decorators must be positioned under a slash_command decorator")

        if not hasattr(func, "permissions"):
            func.permissions = {}

        if guild_id not in func.permissions:
            func.permissions[guild_id] = []
        func.permissions[guild_id] += permissions
        return func

    return wrapper



from typing import TYPE_CHECKING, Any, List, Optional

from dis_snek.const import MISSING
from dis_snek.models.route import Route
from dis_snek.utils.converters import timestamp_converter

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type


class ThreadRequests:
    request: Any

    async def join_thread(self, thread_id: "Snowflake_Type") -> None:
        """
        Join a thread.

        parameters:
            thread_id: The thread to join.
        """
        return await self.request(Route("PUT", f"/channels/{thread_id}/thread-members/@me"))

    async def leave_thread(self, thread_id: "Snowflake_Type") -> None:
        """
        Leave a thread.

        parameters:
            thread_id: The thread to leave.
        """
        return await self.request(Route("DELETE", f"/channels/{thread_id}/thread-members/@me"))

    async def add_thread_member(self, thread_id: "Snowflake_Type", user_id: "Snowflake_Type") -> None:
        """
        Add another user to a thread.

        parameters:
            thread_id: The ID of the thread
            user_id: The ID of the user to add
        """
        return await self.request(Route("PUT", f"/channels/{thread_id}/thread-members/{user_id}"))

    async def remove_thread_member(self, thread_id: "Snowflake_Type", user_id: "Snowflake_Type") -> None:
        """
        Remove another user from a thread.

        parameters:
            thread_id: The ID of the thread
            user_id: The ID of the user to remove
        """
        return await self.request(Route("DELETE", f"/channels/{thread_id}/thread-members/{user_id}"))

    async def list_thread_members(self, thread_id: "Snowflake_Type") -> List[dict]:
        """
        Get a list of members in the thread.

        parameters:
            thread_id: the id of the thread
        returns:
            a list of member objects
        """
        return await self.request(Route("GET", f"/channels/{thread_id}/thread-members"))

    async def list_public_archived_threads(
        self, channel_id: "Snowflake_Type", limit: int = None, before: Optional["Snowflake_Type"] = None
    ) -> dict:
        """
        Get a list of archived public threads in a channel.

        parameters:
            channel_id: The channel to get threads from
            limit: Optional limit of threads to
            before: Get threads before this snowflake
        returns:
            a list of threads
        """
        payload = {}
        if limit:
            payload["limit"] = limit
        if before:
            payload["before"] = timestamp_converter(before)
        return await self.request(Route("GET", f"/channels/{channel_id}/threads/archived/public"), params=payload)

    async def list_private_archived_threads(
        self, channel_id: "Snowflake_Type", limit: int = None, before: Optional["Snowflake_Type"] = None
    ) -> dict:
        """
        Get a list of archived private threads in a channel.

        parameters:
            channel_id: The channel to get threads from
            limit: Optional limit of threads to
            before: Get threads before this snowflake
        returns:
            a list of threads
        """
        payload = {}
        if limit:
            payload["limit"] = limit
        if before:
            payload["before"] = before
        return await self.request(Route("GET", f"/channels/{channel_id}/threads/archived/private"), params=payload)

    async def list_joined_private_archived_threads(
        self, channel_id: "Snowflake_Type", limit: int = None, before: Optional["Snowflake_Type"] = None
    ) -> dict:
        """
        Get a list of archived private threads in a channel that you have joined.

        parameters:
            channel_id: The channel to get threads from
            limit: Optional limit of threads to
            before: Get threads before this snowflake
        returns:
            a list of threads
        """
        payload = {}
        if limit:
            payload["limit"] = limit
        if before:
            payload["before"] = before
        return await self.request(
            Route("GET", f"/channels/{channel_id}/users/@me/threads/archived/private"), params=payload
        )

    async def list_active_threads(self, guild_id: "Snowflake_Type") -> dict:
        """
        List active threads within a guild.

        parameters:
            guild_id: the guild id to get threads from
        returns:
            A list of active threads
        """
        return await self.request(Route("GET", f"/guilds/{guild_id}/threads/active"))

    async def create_thread(
        self,
        channel_id: "Snowflake_Type",
        name: str,
        auto_archive_duration: int,
        thread_type: int = None,
        invitable: Optional[bool] = None,
        message_id: Optional["Snowflake_Type"] = None,
        reason: str = MISSING,
    ) -> dict:
        """
        Create a thread in the given channel.
        Can either create a thread with or without a message

        parameters:
            channel_id: The ID of the channel to create this thread in
            name: The name of the thread
            auto_archive_duration: duration in minutes to automatically archive the thread after recent activity,
            can be set to: 60, 1440, 4320, 10080
            thread_type: The type of thread, defaults to public. ignored if creating thread from a message
            invitable:
            message_id: An optional message to create a thread from.
            reason: An optional reason for the audit log
        returns:
            The created thread
        """
        payload = dict(name=name, auto_archive_duration=auto_archive_duration)
        if message_id:
            return await self.request(
                Route("POST", f"/channels/{channel_id}/messages/{message_id}/threads"), data=payload, reason=reason
            )
        else:
            payload["type"] = thread_type
            payload["invitable"] = invitable
            return await self.request(Route("POST", f"/channels/{channel_id}/threads"), data=payload, reason=reason)



"""
This file outlines the interaction between snek and Discord's Gateway API.
"""
import asyncio
import concurrent.futures
import logging
import sys
import threading
import time
import zlib
from typing import Any, Callable, Coroutine, List, Optional

from aiohttp import WSMsgType

from dis_snek.const import logger_name, events, MISSING
from dis_snek.errors import WebSocketClosed, WebSocketRestart
from dis_snek.http_client import DiscordClientWebSocketResponse, HTTPClient
from dis_snek.models.enums import Intents, Status
from dis_snek.models.enums import WebSocketOPCodes as OPCODE
from dis_snek.utils.input_utils import OverriddenJson
from dis_snek.utils.serializer import dict_filter_none

log = logging.getLogger(logger_name)


class BeeGees(threading.Thread):
    """
    Keeps the gateway connection alive.

    ♫ Stayin' Alive ♫

    Parameters:
        ws WebsocketClient: WebsocketClient
        interval int: How often to send heartbeats -- dictated by discord

    Attributes:
        behind_msg: The log message used when the heartbeat is late
        block_msg: The log message used when the heartbeat is blocked
        msg: The log message used when a hearbeat is sent
        last_recv: When the last heartbeat was received
    """

    ws: Any  # actually WebsocketClient, but these 2 classes reference each other so ¯\_(ツ)_/¯
    _main_thread_id: int
    _interval: int
    daemon: bool
    msg: str
    block_msg: str
    behind_msg: str
    _stop_ev: threading.Event
    _last_ack: float
    _last_send: float
    last_recv: float
    heartbeat_timeout: int

    def __init__(self, ws: Any, interval: int) -> None:
        self.ws = ws
        self._main_thread_id = ws.thread_id
        self.interval = interval
        super().__init__()
        self.daemon = True

        self.msg = "Keeping shard ID %s websocket alive with sequence %s."
        self.block_msg = "Shard ID %s heartbeat blocked for more than %s seconds."
        self.behind_msg = "Can't keep up, shard ID %s websocket is %.1fs behind."

        self._stop_ev = threading.Event()
        self._last_ack = time.perf_counter()
        self._last_send = time.perf_counter()
        self._last_recv = time.perf_counter()
        self.latency = float("inf")
        self.heartbeat_timeout = ws._max_heartbeat_timeout

    def run(self) -> None:
        """Start automatically sending heartbeats to discord."""
        while not self._stop_ev.wait(self.interval):

            data = self.get_payload()
            log.debug(self.msg, 0, data["d"])
            f = asyncio.run_coroutine_threadsafe(self.ws.send_heartbeat(data), loop=self.ws.http.loop)
            duration = 0
            try:
                # block until sending is complete
                while True:
                    try:
                        f.result(10)
                        break
                    except concurrent.futures.TimeoutError:
                        duration += 10
                        log.warning("Heartbeat took too long")
            except Exception:
                log.debug(f"Unable to send heartbeat for {duration} seconds. Closing")
                self.stop()
            else:
                self._last_send = time.perf_counter()

    def get_payload(self) -> dict:
        """
        Get a payload representing a heartbeat.

        returns:
            representation of the heartbeat payload
        """
        return {"op": OPCODE.HEARTBEAT, "d": self.ws.sequence}

    def stop(self) -> None:
        """Stop sending heartbeats."""
        self._stop_ev.set()

    def recv(self) -> None:
        """Log the time that the last heartbeat was received."""
        self._last_recv = time.perf_counter()

    def ack(self) -> None:
        """Log discord ack the heartbeat."""
        ack_time = time.perf_counter()
        self._last_ack = time.perf_counter()
        self.latency = ack_time - self._last_send
        if self.latency > 10:
            log.warning(self.behind_msg, 0, self.latency)


class WebsocketClient:
    """
    Manages the connection to discord's websocket.

    Parameters:
        session_id: The session_id to use, if resuming
        sequence: The sequence to use, if resuming

    Attributes:
        buffer: A buffer to hold incoming data until its complete
        compress: Should the connection be compressed
        intents: The intents used in this connection
        sequence: The sequence of this connection
        session_id: The session ID of this connection
    """

    __slots__ = (
        "http",
        "_gateway",
        "ws",
        "intents",
        "session_id",
        "dispatch",
        "sequence",
        "_keep_alive",
        "_closed",
        "buffer",
        "_zlib",
        "_max_heartbeat_timeout",
        "thread_id",
        "_trace",
    )
    buffer: bytearray
    _closed: bool
    compress: int
    dispatch: Callable[..., Coroutine]
    _gateway: str
    http: HTTPClient
    _keep_alive: Optional[BeeGees]
    _max_heartbeat_timeout: int
    intents: Intents
    sequence: Optional[int]
    session_id: Optional[int]
    thread_id: int
    _trace: List[str]
    ws: DiscordClientWebSocketResponse

    def __init__(self, session_id: Optional[int] = None, sequence: Optional[int] = None) -> None:
        self.session_id = session_id
        self.sequence = sequence

        self.buffer = bytearray()
        self._zlib = zlib.decompressobj()
        self._keep_alive = MISSING

        self._max_heartbeat_timeout = 120
        self.thread_id = threading.get_ident()
        self._trace = []

        self._closed = False

    @classmethod
    async def connect(
        cls,
        http: HTTPClient,
        dispatch: Callable[..., Coroutine],
        intents: Intents,
        resume: bool = False,
        session_id: Optional[int] = None,
        sequence: Optional[int] = None,
    ):
        """
        Connect to the discord gateway.

        parameters:
            http: The HTTPClient
            dispatch: A method to dispatch events
            intents: The intents of this bot
            resume: Are we attempting to resume?
            session_id: The session id to use, if resuming
            sequence: The sequence to use, if resuming
        """
        cls.http = http
        cls._gateway = await http.get_gateway()
        cls.intents = intents
        cls.dispatch = dispatch
        cls.ws = await cls.http.websocket_connect(cls._gateway)
        dispatch(events.Connect())
        if resume:
            return cls(session_id, sequence)
        return cls()

    @property
    def latency(self) -> float:
        """Get the latency of the connection."""
        return float("inf") if not self._keep_alive else self._keep_alive.latency

    async def _receive(self) -> None:
        resp = await self.ws.receive()
        msg = resp.data

        if isinstance(resp.data, bytes):
            self.buffer.extend(msg)

            if len(msg) < 4 or msg[-4:] != b"\x00\x00\xff\xff":
                # message isnt complete yet, wait
                return

            msg = self._zlib.decompress(self.buffer)
            self.buffer = bytearray()

            msg = msg.decode("utf-8")

        if resp.type == WSMsgType.CLOSE:
            await self.close(msg)
            return

        if resp.type == WSMsgType.CLOSING:
            return

        if resp.type == WSMsgType.CLOSED:
            raise WebSocketClosed(1000)

        msg = OverriddenJson.loads(msg)
        log.debug(f"Websocket Event: {msg}")

        op = msg.get("op")
        data = msg.get("d")
        seq = msg.get("s")
        if seq is not None:
            self.sequence = seq

        if op != OPCODE.DISPATCH:
            if op == OPCODE.HELLO:
                interval = data["heartbeat_interval"] / 1000
                self._keep_alive = BeeGees(ws=self, interval=interval)
                await self.send_json(self._keep_alive.get_payload())
                self._keep_alive.start()
                if not self.session_id:
                    await self.identify()
                else:
                    await self.resume()
                return

            if op == OPCODE.HEARTBEAT:
                if self._keep_alive:
                    await self.send_json(self._keep_alive.get_payload())
                return
            if op == OPCODE.HEARTBEAT_ACK:
                if self._keep_alive:
                    self._keep_alive.ack()
                return

            if op in (OPCODE.INVALIDATE_SESSION, OPCODE.RECONNECT):
                # session invalidated, restart
                log.debug(f"Reconnecting to discord due to opcode {op}::{OPCODE(op).name}")
                if data is True or op == OPCODE.RECONNECT:
                    await self.close()
                    raise WebSocketRestart(True)

                self.session_id = self.sequence = None
                log.warning("Session has been invalidated")

                await self.close(code=1000)
                raise WebSocketRestart

        else:
            if msg.get("t") == "READY":
                self._trace = data.get("_trace", [])
                self.sequence = msg["s"]
                self.session_id = data["session_id"]
                log.info(f"Successfully connected to Gateway! Trace: {self._trace} Session_ID: {self.session_id}")
                self.dispatch(events.RawGatewayEvent(data, override_name="websocket_ready"))
                return
            else:
                self.dispatch(events.RawGatewayEvent(msg, override_name="raw_socket_receive"))
            self.dispatch(events.RawGatewayEvent(data, override_name=f"raw_{msg.get('t').lower()}"))

    async def run(self) -> None:
        """Start receiving events from the websocket."""
        while not self._closed:
            await self._receive()

    def __del__(self) -> None:
        if not self._closed:
            self.http.loop.run_until_complete(self.close())

    async def close(self, code: int = 1000) -> None:
        """
        Close the connection to the gateway.

        Parameters:
            code: the close code to use
        """
        if self._closed:
            return
        await self.ws.close(code=code)
        if isinstance(self._keep_alive, BeeGees):
            self._keep_alive.stop()
        self._closed = True

    async def send(self, data: str) -> None:
        """
        Send data to the gateway.

        Parameters:
            data: The data to send
        """
        await self.ws.send_str(data)

    async def send_json(self, data: dict) -> None:
        """
        Send json data to the gateway.

        Parameters:
            data: The data to send
        """
        data = OverriddenJson.dumps(data)
        log.debug(f"Sending data to gateway: {data}")
        await self.send(data)

    async def identify(self) -> None:
        """Send an identify payload to the gateway."""
        data = {
            "op": OPCODE.IDENTIFY,
            "d": {
                "token": self.http.token,
                "intents": self.intents,
                "large_threshold": 250,
                "properties": {"$os": sys.platform, "$browser": "dis.snek", "$device": "dis.snek"},
            },
            "compress": True,
        }
        await self.send_json(data)
        log.debug(f"Client has identified itself to Gateway, requesting intents: {self.intents}!")

    async def resume(self) -> None:
        """Send a resume payload to the gateway."""
        data = {
            "op": OPCODE.RESUME,
            "d": {"token": self.http.token, "seq": self.sequence, "session_id": self.session_id},
        }
        await self.send_json(data)
        self.dispatch(events.Resume())
        log.debug("Client is attempting to resume a connection")

    async def send_heartbeat(self, data: dict) -> None:
        """Send a heartbeat to the gateway."""
        await self.send_json(data)

    async def change_presence(self, activity=None, status: Status = Status.ONLINE, since=None):
        payload = dict_filter_none(
            {
                "since": since if since else time.time(),
                "activities": [activity] if activity else None,
                "status": status,
                "afk": status == Status.AFK,
            }
        )
        await self.send_json({"op": OPCODE.PRESENCE, "d": payload})



"""
Constants used throughout Snek.

attributes:
    __version__ str: The version of the library.
    __repo_url__ str: The URL of the repository.
    __py_version__ str: The python version in use.
    logger_name str: The logger name used by Snek.
    kwarg_spam bool: Should ``unused kwargs`` be logged.

    ACTION_ROW_MAX_ITEMS int: The maximum number of items in an action row.
    SELECTS_MAX_OPTIONS int: The maximum number of options a select may have.
    SELECT_MAX_NAME_LENGTH int: The max length of a select's name.

    CONTEXT_MENU_NAME_LENGTH int: The max length of a context menu's name.
    SLASH_CMD_NAME_LENGTH int: The max length of a slash command's name.
    SLASH_CMD_MAX_DESC_LENGTH int: The maximum length of a slash command's description.
    SLASH_CMD_MAX_OPTIONS int: The maximum number of options a slash command may have.
    SLASH_OPTION_NAME_LENGTH int: The maximum length of a slash option's name.

    EMBED_MAX_NAME_LENGTH int: The maximum length for an embed title
    EMBED_MAX_DESC_LENGTH int: The maximum length for an embed description
    EMBED_MAX_FIELDS int: The maximum number of fields for an embed
    EMBED_TOTAL_MAX int: The total combined number of characters for an embed

    GLOBAL_SCOPE int: Represents a global scope for application commands.

    MISSING _sentinel: A sentinel value that indicates something has not been set
"""

import sys
import sentinel
from collections import defaultdict
import dis_snek.models.events as events


_ver_info = sys.version_info


__version__ = "0.0.0"
__repo_url__ = "https://github.com/LordOfPolls/dis_snek"
__py_version__ = f"{_ver_info[0]}.{_ver_info[1]}"
logger_name = "dis.snek"
kwarg_spam = False

DISCORD_EPOCH = 1420070400000

ACTION_ROW_MAX_ITEMS = 5
SELECTS_MAX_OPTIONS = 25
SELECT_MAX_NAME_LENGTH = 100

CONTEXT_MENU_NAME_LENGTH = 32
SLASH_CMD_NAME_LENGTH = 32
SLASH_CMD_MAX_DESC_LENGTH = 100
SLASH_CMD_MAX_OPTIONS = 25
SLASH_OPTION_NAME_LENGTH = 100

EMBED_MAX_NAME_LENGTH = 256
EMBED_MAX_DESC_LENGTH = 4096
EMBED_MAX_FIELDS = 25
EMBED_TOTAL_MAX = 6000

GLOBAL_SCOPE = sentinel.create(
    "GLOBAL_SCOPE",
    (int,),
    cls_dict={
        "__name__": "GLOBAL_SCOPE",
        "__getattr__": lambda x, y: 0,
        "__hash__": lambda _: hash(0),
    },
)

MISSING = sentinel.create(
    "MISSING",
    cls_dict={
        "__eq__": lambda x, y: x.__name__ == y.__name__ if hasattr(y, "__name__") else False,
        "__name__": "MISSING",
        "__getattr__": lambda x, y: None,
        "__bool__": lambda _: False,
    },
)

PREMIUM_GUILD_LIMITS = defaultdict(
    lambda: {"emoji": 50, "stickers": 0, "bitrate": 96000, "filesize": 8388608},
    {
        1: {"emoji": 100, "stickers": 15, "bitrate": 128000, "filesize": 8388608},
        2: {"emoji": 150, "stickers": 30, "bitrate": 256000, "filesize": 52428800},
        3: {"emoji": 250, "stickers": 60, "bitrate": 384000, "filesize": 104857600},
    },
)
"""Limits granted per premium level of a guild"""



import asyncio
import time
from asyncio import QueueEmpty
from collections import namedtuple
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union, Callable

import attr
from attr.converters import optional as optional_c

from dis_snek.const import MISSING, DISCORD_EPOCH
from dis_snek.mixins.send import SendMixin
from dis_snek.models.discord import DiscordObject
from dis_snek.models.discord_objects.invite import Invite, InviteTargetTypes
from dis_snek.models.discord_objects.thread import ThreadMember, ThreadList
from dis_snek.models.enums import ChannelTypes, OverwriteTypes, Permissions, VideoQualityModes, AutoArchiveDuration
from dis_snek.models.snowflake import SnowflakeObject, to_snowflake
from dis_snek.models.timestamp import Timestamp
from dis_snek.utils.attr_utils import define, field
from dis_snek.utils.converters import timestamp_converter
from dis_snek.utils.serializer import to_image_data

if TYPE_CHECKING:
    from io import IOBase
    from pathlib import Path

    from aiohttp import FormData

    from dis_snek.client import Snake
    from dis_snek.models.discord_objects.application import Application
    from dis_snek.models.discord_objects.role import Role
    from dis_snek.models.discord_objects.message import Message
    from dis_snek.models.discord_objects.user import User, Member
    from dis_snek.models.snowflake import Snowflake_Type


class ChannelHistory(AsyncIterator):
    """
    An async iterator for searching through a channel's history
    Args:
        channel: The channel to search through
        limit: The maximum number of messages to return (set to 0 for no limit)
        before: get messages before this message ID
        after: get messages after this message ID
        around: get messages "around" this message ID
    """

    # todo: Maybe a async iterators baseclass, most of this behaviour is reusable
    def __init__(self, channel: "BaseChannel", limit=50, before=None, after=None, around=None):

        self.messages = asyncio.Queue()
        self.channel: "BaseChannel" = channel
        self._limit: int = limit if limit else MISSING
        self._last: "Message" = MISSING
        self._retrieved: int = 0
        self._retrieved_messages: list["Message"] = []
        self.before: "Snowflake_Type" = before
        self.after: "Snowflake_Type" = after
        self.around: "Snowflake_Type" = around

    @property
    def _continue(self):
        if not self._limit:
            return True
        return not self._retrieved >= self._limit

    @property
    def _get_limit(self):
        return min(self._limit - self._retrieved, 100) if self._limit else 100

    async def _fetch_messages(self):
        if self._continue:
            messages = []

            if self.after:
                if not self._last:
                    self._last = namedtuple("temp", "id")
                    self._last.id = self.after
                messages = await self.channel.get_messages(limit=self._get_limit, after=self._last.id)
                messages.sort(key=lambda x: x.id)

            elif self.around:
                messages = await self.channel.get_messages(limit=self._get_limit, around=self.around)
                # todo: decide how getting *more* messages from `around` would work
                self._limit = 1  # stops history from getting more messages

            else:
                if self.before and not self._last:
                    self._last = namedtuple("temp", "id")
                    self._last.id = self.before

                messages = await self.channel.get_messages(limit=self._get_limit, before=self._last.id)
                messages.sort(key=lambda x: x.id, reverse=True)

            self._retrieved += len(messages)
            [await self.messages.put(m) for m in messages]

        else:
            raise QueueEmpty()

    async def __anext__(self):
        try:
            if self.messages.empty():
                await self._fetch_messages()
            self._last = self.messages.get_nowait()

            # add the message to the already retrieved messages, so that the search function works when calling it multiple times
            self._retrieved_messages.append(self._last)

            return self._last
        except QueueEmpty:
            raise StopAsyncIteration()

    async def flatten(self):
        """Flatten this iterator into a list of objects"""
        return [elem async for elem in self]

    async def search(self, message_id: "Snowflake_Type") -> bool:
        """
        Check if a message_id exists in the specified history.

        parameters:
            message_id: ID of message to check.
        """

        message_id = int(message_id)

        # check if in the message_id was already retrieved
        if message_id in [message.id for message in self._retrieved_messages]:
            return True

        # loop through remaining history
        async for message in self:
            if message.id == message_id:
                return True

        return False


@define()
class PermissionOverwrite(SnowflakeObject):
    # Todo: Document
    type: "OverwriteTypes" = field(repr=True, converter=OverwriteTypes)
    allow: "Permissions" = field(repr=True, converter=Permissions)
    deny: "Permissions" = field(repr=True, converter=Permissions)


@define(slots=False)
class MessageableMixin(SendMixin):
    last_message_id: Optional["Snowflake_Type"] = attr.ib(
        default=None
    )  # TODO May need to think of dynamically updating this.
    default_auto_archive_duration: int = attr.ib(default=AutoArchiveDuration.ONE_DAY)
    last_pin_timestamp: Optional[Timestamp] = attr.ib(default=None, converter=optional_c(timestamp_converter))

    async def _send_http_request(self, message_payload: Union[dict, "FormData"]) -> dict:
        return await self._client.http.create_message(message_payload, self.id)

    async def get_message(self, message_id: "Snowflake_Type") -> "Message":
        """
        Fetch a message from the channel.

        parameters:
            message_id: ID of message to retrieve.

        returns:
            The message object fetched.
        """
        message_id = to_snowflake(message_id)
        message: "Message" = await self._client.cache.get_message(self.id, message_id)
        return message

    def history(
        self,
        limit=100,
        before: "Snowflake_Type" = None,
        after: "Snowflake_Type" = None,
        around: "Snowflake_Type" = None,
    ):
        """
        Get an async iterator for the history of this channel
        Args:
            limit: The maximum number of messages to return (set to 0 for no limit)
            before: get messages before this message ID
            after: get messages after this message ID
            around: get messages "around" this message ID

        ??? Hint "Example Usage:"
            ```python
            async for message in channel.history(limit=0):
                if message.author.id == 174918559539920897:
                    print("Found author's message")
                    # ...
                    break
            ```
            or
            ```python
            history = channel.history(limit=250)
            # Flatten the async iterator into a list
            messages = await history.flatten()
            ```

        Returns:
            ChannelHistory (AsyncIterator)
        """
        return ChannelHistory(self, limit, before, after, around)

    async def get_messages(
        self,
        limit: int = 50,
        around: "Snowflake_Type" = MISSING,
        before: "Snowflake_Type" = MISSING,
        after: "Snowflake_Type" = MISSING,
    ) -> List["Message"]:
        """
        Fetch multiple messages from the channel.

        parameters:
            limit: Max number of messages to return, default `50`, max `100`
            around: Message to get messages around
            before: Message to get messages before
            after: Message to get messages after

        returns:
            A list of messages fetched.
        """
        if limit > 100:
            raise ValueError("You cannot fetch more than 100 messages at once.")

        if around:
            around = to_snowflake(around)
        elif before:
            before = to_snowflake(before)
        elif after:
            after = to_snowflake(after)

        messages_data = await self._client.http.get_channel_messages(self.id, limit, around, before, after)
        return [self._client.cache.place_message_data(message_data) for message_data in messages_data]

    async def get_pinned_messages(self) -> List["Message"]:
        """
        Fetch pinned messages from the channel.

        returns:
            A list of messages fetched.
        """
        messages_data = await self._client.http.get_pinned_messages(self.id)
        return [self._client.cache.place_message_data(message_data) for message_data in messages_data]

    async def delete_messages(
        self, messages: List[Union["Snowflake_Type", "Message"]], reason: Optional[str] = MISSING
    ) -> None:
        """
        Bulk delete messages from channel.

        parameters:
            messages: List of messages or message IDs to delete.
            reason: The reason for this action. Used for audit logs.
        """
        message_ids = [to_snowflake(message) for message in messages]
        # TODO Add check for min/max and duplicates.

        if len(message_ids) == 1:
            # bulk delete messages will throw a http error if only 1 message is passed
            await self.delete_message(message_ids[0], reason)
        else:
            await self._client.http.bulk_delete_messages(self.id, message_ids, reason)

    async def delete_message(self, message: Union["Snowflake_Type", "Message"], reason: str = None) -> None:
        """
        Delete a single message from a channel.

        Args:
            message: The message to delete
            reason: The reason for this action
        """
        message = to_snowflake(message)
        await self._client.http.delete_message(self.id, message, reason=reason)

    async def purge(
        self,
        deletion_limit: int = 50,
        search_limit: int = 100,
        predicate: Callable[["Message"], bool] = MISSING,
        before: Optional["Snowflake_Type"] = MISSING,
        after: Optional["Snowflake_Type"] = MISSING,
        around: Optional["Snowflake_Type"] = MISSING,
        reason: Optional[str] = MISSING,
    ) -> int:
        """
        Bulk delete messages within a channel. If a `predicate` is provided, it will be used to determine which messages to delete,
        otherwise all messages will be deleted within the `deletion_limit`

        ??? Hint "Example Usage:"
            ```python
            # this will delete the last 20 messages sent by a user with the given ID
            deleted = await channel.purge(deletion_limit=20, predicate=lambda m: m.author.id == 174918559539920897)
            await channel.send(f"{deleted} messages deleted")
            ```

        Args:
            deletion_limit: The target amount of messages to delete
            search_limit: How many messages to search through
            predicate: A function that returns True or False, and takes a message as an argument
            before: Search messages before this ID
            after: Search messages after this ID
            around: Search messages around this ID
            reason: The reason for this deletion

        Returns:
            The total amount of messages deleted
        """
        if not predicate:
            predicate = lambda m: True  # noqa

        to_delete = []

        # 1209600 14 days ago in seconds, 1420070400000 is used to convert to snowflake
        fourteen_days_ago = int((time.time() - 1209600) * 1000.0 - DISCORD_EPOCH) << 22
        async for message in self.history(limit=search_limit, before=before, after=after, around=around):
            if deletion_limit != 0 and len(to_delete) == deletion_limit:
                break

            if not predicate(message):
                # fails predicate
                continue

            if message.id < fourteen_days_ago:
                # message is too old to be purged
                continue

            to_delete.append(message.id)

        count = len(to_delete)
        while len(to_delete):
            iteration = [to_delete.pop() for i in range(min(100, len(to_delete)))]
            await self.delete_messages(iteration, reason=reason)
        return count

    async def trigger_typing(self) -> None:
        await self._client.http.trigger_typing_indicator(self.id)


@define(slots=False)
class InvitableMixin:
    async def create_invite(
        self,
        max_age: int = 86400,
        max_uses: int = 0,
        temporary: bool = False,
        unique: bool = False,
        target_type: Optional["InviteTargetTypes"] = None,
        target_user: Optional[Union["Snowflake_Type", "User"]] = None,
        target_application: Optional[Union["Snowflake_Type", "Application"]] = None,
        reason: Optional[str] = None,
    ) -> "Invite":
        """
        Create channel invite.

        parameters:
            max_age: Max age of invite in seconds, default 86400 (24 hours).
            max_uses: Max uses of invite, default 0.
            temporary: Grants temporary membership, default False.
            unique: Invite is unique, default false.
            target_type: Target type for streams and embedded applications.
            target_user_id: Target User ID for Stream target type.
            target_application_id: Target Application ID for Embedded App target type.
            reason: The reason for creating this invite.

        returns:
            Newly created Invite object.
        """
        if target_type:
            if target_type == InviteTargetTypes.STREAM and not target_user:
                raise ValueError("Stream target must include target user id.")
            elif target_type == InviteTargetTypes.EMBEDDED_APPLICATION and not target_application:
                raise ValueError("Embedded Application target must include target application id.")

        if target_user and target_application:
            raise ValueError("Invite target must be either User or Embedded Application, not both.")
        elif target_user:
            target_user = to_snowflake(target_user)
            target_type = InviteTargetTypes.STREAM
        elif target_application:
            target_application = to_snowflake(target_application)
            target_type = InviteTargetTypes.EMBEDDED_APPLICATION

        invite_data = await self._client.http.create_channel_invite(
            self.id, max_age, max_uses, temporary, unique, target_type, target_user, target_application, reason
        )
        return Invite.from_dict(invite_data, self._client)

    async def get_invites(self) -> List["Invite"]:
        """
        Gets all invites (with invite metadata) for the channel.
        """
        invites_data = await self._client.http.get_channel_invites(self.id)
        return Invite.from_list(invites_data, self._client)


@define(slots=False)
class ThreadableMixin:
    async def create_thread_with_message(
        self,
        name: str,
        message: Union["Snowflake_Type", "Message"],
        auto_archive_duration: Union[AutoArchiveDuration, int] = AutoArchiveDuration.ONE_DAY,
        reason: Optional[str] = None,
    ) -> Union["GuildNewsThread", "GuildPublicThread"]:
        thread_data = await self._client.http.create_thread(
            channel_id=self.id,
            name=name,
            auto_archive_duration=auto_archive_duration,
            message_id=to_snowflake(message),
            reason=reason,
        )
        return self._client.cache.place_channel_data(thread_data)

    async def create_thread_without_message(
        self,
        name: str,
        thread_type: Union[ChannelTypes, int],
        invitable: Optional[bool] = None,
        auto_archive_duration: Union[AutoArchiveDuration, int] = AutoArchiveDuration.ONE_DAY,
        reason: Optional[str] = None,
    ) -> Union["GuildPrivateThread", "GuildPublicThread"]:
        thread_data = await self._client.http.create_thread(
            channel_id=self.id,
            name=name,
            thread_type=thread_type,
            auto_archive_duration=auto_archive_duration,
            invitable=invitable,
            reason=reason,
        )
        return self._client.cache.place_channel_data(thread_data)

    async def get_public_archived_threads(self, limit: int = None, before: Union["Timestamp"] = None) -> ThreadList:
        threads_data = await self._client.http.list_public_archived_threads(
            channel_id=self.id, limit=limit, before=before
        )
        threads_data["id"] = self.id
        return ThreadList.from_dict(threads_data, self._client)

    async def get_private_archived_threads(self, limit: int = None, before: Union["Timestamp"] = None) -> ThreadList:
        threads_data = await self._client.http.list_private_archived_threads(
            channel_id=self.id, limit=limit, before=before
        )
        threads_data["id"] = self.id
        return ThreadList.from_dict(threads_data, self._client)

    async def get_joined_private_archived_threads(
        self, limit: int = None, before: Union["Timestamp"] = None
    ) -> ThreadList:
        threads_data = await self._client.http.list_joined_private_archived_threads(
            channel_id=self.id, limit=limit, before=before
        )
        threads_data["id"] = self.id
        return ThreadList.from_dict(threads_data, self._client)


@define(slots=False)
class BaseChannel(DiscordObject):
    name: Optional[str] = field(default=None)
    type: Union[ChannelTypes, int] = field(converter=ChannelTypes)

    @classmethod
    def from_dict_factory(cls, data: dict, client: "Snake") -> "TYPE_ALL_CHANNEL":
        """
        Creates a channel object of the appropriate type

        parameters:
            data: The channel data.
            client: The bot.

        returns:
            The new channel object.
        """
        channel_type = data.get("type", None)
        channel_class = TYPE_CHANNEL_MAPPING.get(channel_type, None)
        if not channel_class:
            raise TypeError(f"Unsupported channel type for {data} ({channel_type}), please consult the docs.")

        return channel_class.from_dict(data, client)

    async def _edit(self, payload: dict, reason: Optional[str] = MISSING) -> None:
        """
        # TODO

        Args:
            payload:
            reason:

        Returns:

        """
        channel_data = self._client.http.modify_channel(self.id, payload, reason)
        self.update_from_dict(channel_data)

    async def delete(self, reason: Optional[str] = MISSING) -> None:
        """
        Delete this channel.

        parameters:
            reason: The reason for deleting this channel
        """
        await self._client.http.delete_channel(self.id, reason)


################################################################
# DMs


@define()
class DMChannel(BaseChannel, MessageableMixin):
    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        data = super()._process_dict(data, client)
        data["recipients"] = [client.cache.place_user_data(recipient) for recipient in data["recipients"]]
        return data

    async def edit(
        self,
        name: Optional[str] = MISSING,
        icon: Optional[Union[str, "Path", "IOBase"]] = MISSING,
        reason: Optional[str] = MISSING,
    ):
        payload = dict(name=name, icon=to_image_data(icon))
        await self._edit(payload=payload, reason=reason)


@define()
class DM(DMChannel):
    recipient: "User" = field()

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        data = super()._process_dict(data, client)
        data["recipient"] = data.pop("recipients")[0]
        client.cache.place_dm_channel_id(data["recipient"], data["id"])
        return data


@define()
class DMGroup(DMChannel):
    owner_id: "Snowflake_Type" = attr.ib()
    application_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    recipients: List["User"] = field(factory=list)

    async def get_owner(self) -> "User":
        return await self._client.cache.get_user(self.owner_id)

    async def add_recipient(
        self, user: Union["User", "Snowflake_Type"], access_token: str, nickname: Optional[str] = MISSING
    ):
        user = await self._client.cache.get_user(user)
        await self._client.http.group_dm_add_recipient(self.id, user.id, access_token, nickname)
        self.recipients.append(user)

    async def remove_recipient(self, user: Union["User", "Snowflake_Type"]):
        user = await self._client.cache.get_user(user)
        await self._client.http.group_dm_remove_recipient(self.id, user.id)
        self.recipients.remove(user)


################################################################
# Guild


@define()
class GuildChannel(BaseChannel):
    position: Optional[int] = attr.ib(default=0)
    nsfw: bool = attr.ib(default=False)
    parent_id: Optional["Snowflake_Type"] = attr.ib(default=None, converter=optional_c(to_snowflake))

    _guild_id: Optional["Snowflake_Type"] = attr.ib(default=None, converter=optional_c(to_snowflake))
    _permission_overwrites: Dict["Snowflake_Type", "PermissionOverwrite"] = attr.ib(factory=list)

    @property
    def guild(self):
        return self._client.cache.guild_cache.get(self._guild_id)

    @property
    def category(self):
        return self._client.cache.channel_cache.get(self.parent_id)

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        permission_overwrites = data.get("permission_overwrites", [])
        data["permission_overwrites"] = {
            obj.id: obj for obj in (PermissionOverwrite(**permission) for permission in permission_overwrites)
        }
        return data

    async def edit_permission(self, overwrite: PermissionOverwrite, reason: Optional[str] = None) -> None:
        await self._client.http.edit_channel_permission(
            self.id, overwrite.id, overwrite.allow, overwrite.deny, overwrite.type, reason  # TODO Convert to str...?
        )

    async def delete_permission(
        self, target: Union["PermissionOverwrite", "Role", "User"], reason: Optional[str] = MISSING
    ):
        target = to_snowflake(target)
        await self._client.http.delete_channel_permission(self.id, target, reason)


@define()
class GuildCategory(GuildChannel):
    async def edit(self, name, position, permission_overwrites, reason):
        payload = dict(  # TODO Proper processing
            name=name,
            position=position,
            permission_overwrites=permission_overwrites,
        )
        await self._edit(payload=payload, reason=reason)


@define()
class GuildStore(GuildChannel):
    async def edit(self, name, position, permission_overwrites, parent_id, nsfw, reason):
        payload = dict(  # TODO Proper processing
            name=name,
            position=position,
            permission_overwrites=permission_overwrites,
            parent_id=parent_id,
            nsfw=nsfw,
        )
        await self._edit(payload=payload, reason=reason)


@define()
class GuildNews(GuildChannel, MessageableMixin, InvitableMixin, ThreadableMixin):
    topic: Optional[str] = attr.ib(default=None)

    async def edit(
        self,
        name,
        position,
        permission_overwrites,
        parent_id,
        nsfw,
        topic,
        channel_type,
        default_auto_archive_duration,
        reason,
    ):
        payload = dict(  # TODO Proper processing
            name=name,
            position=position,
            permission_overwrites=permission_overwrites,
            parent_id=parent_id,
            nsfw=nsfw,
            topic=topic,
            channel_type=channel_type,
            default_auto_archive_duration=default_auto_archive_duration,
        )
        await self._edit(payload=payload, reason=reason)

    async def follow(self, webhook_channel_id: "Snowflake_Type"):
        await self._client.http.follow_news_channel(self.id, webhook_channel_id)


@define()
class GuildText(GuildNews):
    rate_limit_per_user: int = attr.ib(default=0)

    # noinspection PyMethodOverriding
    async def edit(
        self,
        name,
        position,
        permission_overwrites,
        parent_id,
        nsfw,
        topic,
        channel_type,
        default_auto_archive_duration,
        rate_limit_per_user,
        reason,
    ):
        payload = dict(  # TODO Proper processing
            name=name,
            position=position,
            permission_overwrites=permission_overwrites,
            parent_id=parent_id,
            nsfw=nsfw,
            topic=topic,
            channel_type=channel_type,
            rate_limit_per_user=rate_limit_per_user,
            default_auto_archive_duration=default_auto_archive_duration,
        )
        await self._edit(payload=payload, reason=reason)


################################################################
# Guild Threads


@define()
class ThreadChannel(GuildChannel, MessageableMixin):
    owner_id: "Snowflake_Type" = attr.ib(default=None)
    topic: Optional[str] = attr.ib(default=None)
    message_count: int = attr.ib(default=0)
    member_count: int = attr.ib(default=0)
    archived: bool = attr.ib(default=False)
    auto_archive_duration: int = attr.ib(
        default=attr.Factory(lambda self: self.default_auto_archive_duration, takes_self=True)
    )
    locked: bool = attr.ib(default=False)
    archive_timestamp: Optional[Timestamp] = attr.ib(default=None, converter=optional_c(timestamp_converter))

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        data = super()._process_dict(data, client)
        thread_metadata: dict = data.get("thread_metadata", {})
        data.update(thread_metadata)
        return data

    @property
    def is_private(self) -> bool:
        return self.type == ChannelTypes.GUILD_PRIVATE_THREAD

    @property
    def mention(self) -> str:
        return f"<#{self.id}>"

    async def get_members(self) -> List["ThreadMember"]:
        members_data = await self._client.http.list_thread_members(self.id)
        members = []
        for member_data in members_data:
            members.append(ThreadMember.from_dict(member_data, self._client))
        return members

    async def add_member(self, member: Union["Member", "Snowflake_Type"]) -> None:
        await self._client.http.add_thread_member(self.id, to_snowflake(member))

    async def remove_member(self, member: Union["Member", "Snowflake_Type"]) -> None:
        await self._client.http.remove_thread_member(self.id, to_snowflake(member))

    async def join(self):
        await self._client.http.join_thread(self.id)

    async def leave(self):
        await self._client.http.leave_thread(self.id)


@define()
class GuildNewsThread(ThreadChannel):
    async def edit(self, name, archived, auto_archive_duration, locked, rate_limit_per_user, reason):
        payload = dict(  # TODO Proper processing
            name=name,
            archived=archived,
            auto_archive_duration=auto_archive_duration,
            locked=locked,
            rate_limit_per_user=rate_limit_per_user,
        )
        await self._edit(payload=payload, reason=reason)


@define()
class GuildPublicThread(ThreadChannel):
    async def edit(self, name, archived, auto_archive_duration, locked, rate_limit_per_user, reason):
        payload = dict(  # TODO Proper processing
            name=name,
            archived=archived,
            auto_archive_duration=auto_archive_duration,
            locked=locked,
            rate_limit_per_user=rate_limit_per_user,
        )
        await self._edit(payload=payload, reason=reason)


@define()
class GuildPrivateThread(ThreadChannel):
    invitable: bool = field(default=False)

    async def edit(self, name, archived, auto_archive_duration, locked, rate_limit_per_user, invitable, reason):
        payload = dict(  # TODO Proper processing
            name=name,
            archived=archived,
            auto_archive_duration=auto_archive_duration,
            locked=locked,
            rate_limit_per_user=rate_limit_per_user,
            invitable=invitable,
        )
        await self._edit(payload=payload, reason=reason)


################################################################
# Guild Voices


@define()
class VoiceChannel(GuildChannel):  # TODO May not be needed, can be directly just GuildVoice.
    bitrate: int = attr.ib()
    user_limit: int = attr.ib()
    rtc_region: str = attr.ib(default="auto")
    video_quality_mode: Union[VideoQualityModes, int] = attr.ib(default=VideoQualityModes.AUTO)

    async def edit(
        self,
        name,
        position,
        permission_overwrites,
        parent_id,
        bitrate,
        user_limit,
        rtc_region,
        video_quality_mode,
        reason,
    ):
        payload = dict(  # TODO Proper processing
            name=name,
            position=position,
            permission_overwrites=permission_overwrites,
            parent_id=parent_id,
            bitrate=bitrate,
            user_limit=user_limit,
            rtc_region=rtc_region,
            video_quality_mode=video_quality_mode,
        )
        await self._edit(payload=payload, reason=reason)


@define()
class GuildVoice(VoiceChannel, InvitableMixin):
    pass


@define()
class GuildStageVoice(GuildVoice):
    pass


TYPE_ALL_CHANNEL = Union[
    GuildText,
    GuildNews,
    GuildVoice,
    GuildStageVoice,
    GuildCategory,
    GuildStore,
    GuildPublicThread,
    GuildPrivateThread,
    GuildNewsThread,
    DMChannel,
    DMGroup,
]


TYPE_DM_CHANNEL = Union[DM, DMGroup]


TYPE_GUILD_CHANNEL = Union[GuildCategory, GuildStore, GuildNews, GuildText, GuildVoice, GuildStageVoice]


TYPE_THREAD_CHANNEL = Union[GuildNewsThread, GuildPublicThread, GuildPrivateThread]


TYPE_VOICE_CHANNEL = Union[GuildVoice, GuildStageVoice]


TYPE_MESSAGEABLE_CHANNEL = Union[DM, DMGroup, GuildNews, GuildText, ThreadChannel]


TYPE_CHANNEL_MAPPING = {
    ChannelTypes.GUILD_TEXT: GuildText,
    ChannelTypes.GUILD_NEWS: GuildNews,
    ChannelTypes.GUILD_VOICE: GuildVoice,
    ChannelTypes.GUILD_STAGE_VOICE: GuildStageVoice,
    ChannelTypes.GUILD_CATEGORY: GuildCategory,
    ChannelTypes.GUILD_STORE: GuildStore,
    ChannelTypes.GUILD_PUBLIC_THREAD: GuildPublicThread,
    ChannelTypes.GUILD_PRIVATE_THREAD: GuildPrivateThread,
    ChannelTypes.GUILD_NEWS_THREAD: GuildNewsThread,
    ChannelTypes.DM: DM,
    ChannelTypes.GROUP_DM: DMGroup,
}



from typing import TYPE_CHECKING

import attr

from dis_snek.models.discord import DiscordObject
from dis_snek.utils.attr_utils import define

if TYPE_CHECKING:
    from dis_snek.models.discord_objects.emoji import Emoji


@define()
class Reaction(DiscordObject):
    count: int = attr.ib()
    me: bool = attr.ib(default=False)
    emoji: "Emoji" = attr.ib()

    # TODO: clear, remove, users
    # TODO: http endpoints for reactions require both a channel and message ID parameter, this object lacks both



from typing import TYPE_CHECKING, Any, ClassVar, Optional
from urllib.parse import quote as _uriquote

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type


class Route:
    BASE: ClassVar[str] = "https://discord.com/api/v9"

    def __init__(self, method: str, path: str, **parameters: Any):
        # debug code todo: remove this when no longer needed
        if path[0] != "/":
            print(f"Potentially incorrect path passed: {path}")

        self.path: str = path
        self.method: str = method

        url = f"{self.BASE}{self.path}"
        if parameters:
            url = url.format_map({k: _uriquote(v) if isinstance(v, str) else v for k, v in parameters.items()})
        self.url: str = url

        self.channel_id: Optional["Snowflake_Type"] = parameters.get("channel_id")
        self.guild_id: Optional["Snowflake_Type"] = parameters.get("guild_id")
        self.webhook_id: Optional["Snowflake_Type"] = parameters.get("webhook_id")
        self.webhook_token: Optional[str] = parameters.get("webhook_token")

    @property
    def rl_bucket(self):
        return f"{self.channel_id}:{self.guild_id}:{self.path}"



from typing import Any, List

from dis_snek.models.route import Route


class BotRequests:
    request: Any

    async def get_current_bot_information(self) -> dict:
        """
        Returns the bot's application object without flags.

        returns:
            application object
        """
        return await self.request(Route("GET", "/oauth2/applications/@me"))

    async def get_current_authorisation_information(self) -> dict:
        """
        Returns info about the current authorization

        returns:
            Authorisation information
        """
        return await self.request(Route("GET", "/oauth2/@me"))

    async def list_voice_regions(self) -> List[dict]:
        """
        Returns an array of voice region objects that can be used when setting a voice or stage channel's `rtc_region`.

        returns:
            an array of voice region objects
        """
        return await self.request(Route("GET", "/voice/regions"))



from typing import TYPE_CHECKING, Any, Dict, List

from dis_snek.const import GLOBAL_SCOPE
from dis_snek.models.route import Route

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type


class InteractionRequests:
    request: Any

    async def get_interaction_element(self, application_id: "Snowflake_Type", guild_id: "Snowflake_Type") -> List[Dict]:
        """
        Get all interaction elements for this application from discord.

        parameters:
            application_id: the what application to query
            guild_id: specify a guild to get commands from
        returns:
            InteractionCommand
        """
        if guild_id == GLOBAL_SCOPE:
            return await self.request(Route("GET", f"/applications/{application_id}/commands"))
        return await self.request(Route("GET", f"/applications/{application_id}/guilds/{guild_id}/commands"))

    async def post_interaction_element(
        self, app_id: "Snowflake_Type", data: List[Dict], guild_id: "Snowflake_Type" = None
    ):
        """
        Register an interaction element.

        parameters:
            app_id: The application ID of this bot
            guild_id: The ID of the guild this command is for, if this is a guild command
            data: List of your interaction data
        """

        endpoint = f"/applications/{app_id}/commands"
        if guild_id != GLOBAL_SCOPE:
            endpoint = f"/applications/{app_id}/guilds/{guild_id}/commands"

        return await self.request(Route("PUT", endpoint), data=data)

    async def post_initial_response(self, payload: dict, interaction_id: str, token: str) -> None:
        """
        Post an initial response to an interaction.

        parameters:
            payload: the payload to send
            interaction_id: the id of the interaction
            token: the token of the interaction
        """

        return await self.request(Route("POST", f"/interactions/{interaction_id}/{token}/callback"), data=payload)

    async def post_followup(self, payload: dict, application_id: "Snowflake_Type", token: str) -> None:
        """
        Send a followup to an interaction.

        parameters:
            payload: the payload to send
            application_id: the id of the application
            token: the token of the interaction
        """

        return await self.request(Route("POST", f"/webhooks/{application_id}/{token}"), data=payload)

    async def edit_interaction_message(
        self, payload: dict, application_id: "Snowflake_Type", token: str, message_id: str = "@original"
    ) -> dict:
        """
        Edits an existing interaction message.

        parameters:
            payload: The payload to send.
            application_id: The id of the application.
            token: The token of the interaction.
            message_id: The target message to edit. Defaults to @original which represents the initial response message.

        returns:
            The edited message data.
        """

        return await self.request(
            Route("PATCH", f"/webhooks/{application_id}/{token}/messages/{message_id}"), data=payload
        )

    async def get_interaction_message(self, application_id: str, token: str, message_id: str = "@original") -> dict:
        """
        Gets an existing interaction message.

        parameters:
            payload: The payload to send.
            application_id: The id of the application.
            token: The token of the interaction.
            message_id: The target message to get. Defaults to @original which represents the initial response message.

        returns:
            The message data.
        """

        return await self.request(Route("GET", f"/webhooks/{application_id}/{token}/messages/{message_id}"))

    async def edit_application_command_permissions(
        self,
        application_id: "Snowflake_Type",
        scope: "Snowflake_Type",
        cmd_id: "Snowflake_Type",
        permissions: List[dict],
    ) -> dict:
        """
        Edits command permissions for a specific command.

        parameters:
            application_id: the id of the application
            scope: The scope this command is in
            cmd_id: The command id to edit
            permissions: The permissions to set to this command
        returns:
            Guild Application Command Permissions
        """
        return await self.request(
            Route("PUT", f"/applications/{application_id}/guilds/{scope}/commands/{cmd_id}/permissions"),
            data=permissions,
        )

    async def batch_edit_application_command_permissions(
        self, application_id: "Snowflake_Type", scope: "Snowflake_Type", data: List[dict]
    ) -> dict:
        """
        Edit multiple command permissions within a single scope.

        parameters:
            application_id: the id of the application
            scope: The scope this command is in
            data: The permissions to be set
        returns:
            array of GuildApplicationCommandPermissions objects
        """
        return await self.request(
            Route("PUT", f"/applications/{application_id}/guilds/{scope}/commands/permissions"),
            data=data,
        )

    async def get_application_command_permissions(
        self, application_id: "Snowflake_Type", scope: "Snowflake_Type", cmd_id: "Snowflake_Type"
    ) -> dict:
        """
        Get permission data for a command.

        parameters:
            application_id: the id of the application
            scope: The scope this command is in
            cmd_id: The command id to edit
        returns:
            guild application command permissions
        """
        return await self.request(
            Route("GET", f"/applications/{application_id}/guilds/{scope}/commands/{cmd_id}/permissions")
        )

    async def batch_get_application_command_permissions(
        self, application_id: "Snowflake_Type", scope: "Snowflake_Type"
    ) -> dict:
        """
        Get permission data for all commands in a scope

        parameters:
            application_id: the id of the application
            scope: The scope this command is in
        returns:
            list of guild application command permissions
        """
        return await self.request(Route("GET", f"/applications/{application_id}/guilds/{scope}/commands/permissions"))



from typing import TYPE_CHECKING, List, Dict, Any, Union

from attr.converters import optional

from dis_snek.mixins.send import SendMixin
from dis_snek.models.discord import DiscordObject
from dis_snek.models.snowflake import to_snowflake
from dis_snek.models.timestamp import Timestamp
from dis_snek.utils.attr_utils import define, field
from dis_snek.utils.converters import timestamp_converter

if TYPE_CHECKING:
    from aiohttp import FormData

    from dis_snek.client import Snake
    from dis_snek.models.discord_objects.user import User
    from dis_snek.models.discord_objects.channel import TYPE_THREAD_CHANNEL
    from dis_snek.models.snowflake import Snowflake_Type


@define()
class ThreadMember(DiscordObject, SendMixin):
    """A thread member is used to indicate whether a user has joined a thread or not."""

    join_timestamp: Timestamp = field(converter=timestamp_converter)
    """The time the current user last joined the thread."""
    flags: int = field()
    """Any user-thread settings, currently only used for notifications."""

    _user_id: "Snowflake_Type" = field(converter=optional(to_snowflake))

    async def get_thread(self) -> "TYPE_THREAD_CHANNEL":
        """
        Gets the thread associated with with this member.

        Returns:
            The thread in question
        """
        return await self._client.get_channel(self.id)

    async def get_user(self) -> "User":
        """
        Get the user associated with this thread member.

        Returns:
            The user object
        """
        return await self._client.get_user(self._user_id)

    async def _send_http_request(self, message_payload: Union[dict, "FormData"]) -> dict:
        dm_id = await self._client.cache.get_dm_channel_id(self._user_id)
        return await self._client.http.create_message(message_payload, dm_id)


@define
class ThreadList(DiscordObject):
    """Represents a list of one or more threads."""

    threads: List["TYPE_THREAD_CHANNEL"] = field(factory=list)  # TODO Reference the cache or store actual object?
    """The active threads."""
    members: List[ThreadMember] = field(factory=list)
    """A thread member object for each returned thread the current user has joined."""
    has_more: bool = field(default=False)
    """Whether there are potentially additional threads that could be returned on a subsequent call."""

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        threads = []
        for thread_data in data["threads"]:
            threads.append(client.cache.place_channel_data(thread_data))
        data["threads"] = threads

        members = []
        for member_data in data["members"]:
            members.append(ThreadMember.from_dict(member_data, client))
        data["members"] = threads

        return data
